from argrelay_api_plugin_server_abstract.DelegatorAbstract import DelegatorAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_api_server_cli.schema_response.InvocationInputSchema import invocation_input_desc
from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_server.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import function_container_ipos_
from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_lib_root.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorError import DelegatorError
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaCustomDataDelegatorError import (
    error_code_,
    error_delegator_custom_data_desc,
    error_message_,
)
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
)


class RelayLineArgsServerRequestHandler(AbstractServerRequestHandler):

    def __init__(
        self,
        local_server: LocalServer,
    ):
        super().__init__(
            local_server = local_server,
        )

    def handle_request(
        self,
        call_ctx: CallContext,
    ) -> dict:
        assert call_ctx.server_action is ServerAction.RelayLineArgs
        self._store_usage_stats_entry(call_ctx)

        self.interpret_command(self.local_server, call_ctx)
        ElapsedTime.measure("after_interpret_command")
        is_error = False
        error_message = ""
        error_code = ClientExitCode.ClientSuccess.value

        # The first envelope (`DataEnvelopeSchema`) is assumed to be of
        # `ReservedEnvelopeClass.class_function` with `FunctionEnvelopeInstanceDataSchema` for its `instance_data`:
        if self.interp_ctx.is_func_found():
            delegator_plugin_instance_id = (
                self
                .interp_ctx
                .envelope_containers[function_container_ipos_]
                .data_envelopes[0][instance_data_][delegator_plugin_instance_id_]
            )
        else:
            is_error = True
            error_message = "ERROR: Function is not selected, try help, or press Tab to complete selection."
            error_code = ClientExitCode.GeneralError.value
            # TODO: TODO_62_75_33_41: Do not hardcode plugin instance id (instance of `DelegatorError`):
            delegator_plugin_instance_id = f"{DelegatorError.__name__}.default"

        delegator_plugin: DelegatorAbstract = self.local_server.server_config.action_delegators[
            delegator_plugin_instance_id
        ]
        invocation_input: InvocationInput = delegator_plugin.run_invoke_control(
            self.interp_ctx,
            self.local_server,
        )

        if is_error:
            invocation_input.custom_plugin_data = {
                error_message_: error_message,
                error_code_: error_code,
            }
            error_delegator_custom_data_desc.validate_dict(invocation_input.custom_plugin_data)

        response_dict = invocation_input_desc.dict_schema.dump(invocation_input)
        return response_dict



from __future__ import annotations

from unittest import mock


class OpenFileMock:
    """
    Mock file data for specific paths:
    https://stackoverflow.com/a/69681105/441652

    Access to file paths not configured within `OpenFileMock` instance is passed through
    to original built-in open func.
    """
    builtin_open = open

    def __init__(self, path_to_data: dict[str, any]):
        self.path_to_data: dict[str, any] = path_to_data
        self.path_to_mock: dict = {}

    def add_path_data(
        self,
        file_path: str,
        file_data: any,
    ) -> None:
        self.path_to_data[file_path] = file_data

    def open(self, *args, **kwargs):
        file_path = args[0]
        if file_path in self.path_to_data:
            new_mock = mock.mock_open(read_data = self.path_to_data[file_path])
            self.path_to_mock[file_path] = new_mock
            return new_mock(*args, **kwargs)
        return self.builtin_open(*args, **kwargs)



import argrelay
from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_lib_check_env_plugin_core.PluginCheckEnvServerResponseValueAbstract import (
    PluginCheckEnvServerResponseValueAbstract,
)
from argrelay_lib_check_env_plugin_core.SchemaPluginCheckEvnServerResponseValueAbstract import (
    field_values_to_command_lines_,
)
from argrelay_lib_root.enum_desc.CheckEnvField import CheckEnvField
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory


class PluginCheckEnvServerResponseValueVersion(PluginCheckEnvServerResponseValueAbstract):
    # TODO: TODO_69_59_78_78: register known files as enum with metadata:
    bootstrap_rel_path = "exe/bootstrap_env.bash"

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict = plugin_config_dict or {
                field_values_to_command_lines_: {
                    CheckEnvField.server_version.name: "argrelay.check_env server_version",
                },
            },
        )

    def verify_online_value(
        self,
        field_name: str,
        field_value,
    ) -> CheckEnvResult:
        client_version = argrelay.__version__

        if client_version == field_value:
            return CheckEnvResult(
                result_category = ResultCategory.VerificationSuccess,
                result_key = field_name,
                result_value = field_value,
                result_message = f"It matches client `argrelay_module_version`",
            )
        else:
            return CheckEnvResult(
                result_category = ResultCategory.VerificationWarning,
                result_key = field_name,
                result_value = field_value,
                result_message = f"It does not match client `argrelay_module_version` => re-base and re-run `@/{self.bootstrap_rel_path}`",
            )



from enum import (
    auto,
    Enum,
)


class CheckEnvField(Enum):
    """
    Field names reported by FS_36_17_84_44 `check_env`.
    """

    server_version = auto()

    server_git_commit_id = auto()

    server_start_time = auto()



from __future__ import annotations

import os
import subprocess
from typing import Union

from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_api_plugin_check_env_abstract.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory
from argrelay_lib_root.enum_desc.TopDir import TopDir
from argrelay_lib_root.misc_helper_common import get_argrelay_dir


class PluginCheckEnvClientLinkedCommands(PluginCheckEnvAbstract):
    """
    This plugin checks which `argrelay_dir` each command is linked to.

    This is related to FS_57_36_37_48 multiple clients coexistence.

    It loops through values keys of `argrelay_basename_to_client_path_map` shell variable
    (which is supposed to map each linked command to a client with potentially different `argrelay_dir`).
    Then, it verifies `argrelay_dir` of the mapped client with the `argrelay_dir` of currently running process.
    """

    # noinspection PyMethodMayBeStatic
    def execute_check(
        self,
        online_mode: Union[bool, None],
    ) -> list[CheckEnvResult]:
        dev_shell_path = os.path.join(
            get_argrelay_dir(),
            TopDir.exe_dir.value,
            "dev_shell.bash",
        )
        shell_var_value = None
        try:
            bash_proc = subprocess.run(
                args = [
                    dev_shell_path,
                    f"echo ${{!argrelay_basename_to_client_path_map[@]}}",
                ],
                capture_output = True,
            )
            assert bash_proc.returncode == 0
            stdout_str = bash_proc.stdout.decode("utf-8")
            shell_var_value = stdout_str.strip()
        except:
            shell_var_value = None

        if shell_var_value is None:
            return [CheckEnvResult(
                result_category = ResultCategory.ExecutionFailure,
                result_key = "client_linked_command",
                result_value = str(shell_var_value),
                result_message = f"Unable to retrieve keys of `argrelay_basename_to_client_path_map` under started: {dev_shell_path}",
            )]
        else:
            linked_commands: list[str] = shell_var_value.split()
            check_env_results: list[CheckEnvResult] = []
            for command_index, linked_command in enumerate(linked_commands):

                shell_var_value = None
                try:
                    bash_proc = subprocess.run(
                        args = [
                            dev_shell_path,
                            f"echo ${{argrelay_basename_to_client_path_map[{linked_command}]}}",
                        ],
                        capture_output = True,
                    )
                    assert bash_proc.returncode == 0
                    stdout_str = bash_proc.stdout.decode("utf-8")
                    shell_var_value = stdout_str.strip()
                except:
                    shell_var_value = None

                if shell_var_value is None:
                    check_env_results.append(CheckEnvResult(
                        result_category = ResultCategory.ExecutionFailure,
                        result_key = f"client_linked_command[{command_index}]",
                        result_value = str(linked_command),
                        result_message = f"Unable to map `{linked_command}` to `@/exe/run_argrelay_client`",
                    ))
                else:
                    argrelay_client_path = shell_var_value
                    if argrelay_client_path == f"{get_argrelay_dir()}/exe/run_argrelay_client":
                        check_env_results.append(CheckEnvResult(
                            result_category = ResultCategory.VerificationSuccess,
                            result_key = f"client_linked_command[{command_index}]",
                            result_value = str(linked_command),
                            result_message = "Command is linked to `@/exe/run_argrelay_client` from this `argrelay_dir`",
                        ))
                    else:
                        check_env_results.append(CheckEnvResult(
                            result_category = ResultCategory.VerificationWarning,
                            result_key = f"client_linked_command[{command_index}]",
                            result_value = str(linked_command),
                            result_message = f"Command is linked to another `argrelay_dir` with: {argrelay_client_path}",
                        ))

            return check_env_results



from __future__ import annotations

from typing import Union

from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_app_server.relay_server.IndexModel import index_props_
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_app_server.runtime_context.SearchControl import SearchControl
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    envelope_payload_,
)
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import (
    populate_search_control,
    search_control_desc,
)

collection_name_container_ipos_ = 1
data_envelope_container_ipos_ = 2


def get_collection_search_control(
) -> dict:
    return populate_search_control(
        ReservedEnvelopeClass.class_collection.name,
        {
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_collection.name,
        },
        [
            # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
            {"class": ReservedPropName.envelope_class.name},

            {"collection": ReservedPropName.collection_name.name},
        ],
    )


class DelegatorDataBackendBase(DelegatorSingleFuncAbstract):
    """
    Implements base functionality for FS_74_69_61_79 get set data envelope.
    """

    def run_search_control(
        self,
        interp_ctx: InterpContext,
        function_data_envelope: dict,
        func_param_container_offset: int,
    ) -> Union[SearchControl, None]:
        """
        Provides `search_control` based on `index_prop` for FS_74_69_61_79 get set data envelope.
        """
        search_control_list: list[SearchControl] = self.extract_search_control_from_function_data_envelope(
            function_data_envelope,
        )
        if func_param_container_offset < len(search_control_list):
            return search_control_list[func_param_container_offset]
        elif func_param_container_offset == len(search_control_list):
            # TODO: TODO_73_23_85_93: use helper to select container ipos:
            collection_name_container = interp_ctx.envelope_containers[
                interp_ctx.curr_interp.base_container_ipos + collection_name_container_ipos_
            ]
            collection_name_container.data_envelopes = (
                interp_ctx
                .query_engine
                .query_data_envelopes_for(collection_name_container)
            )
            if len(collection_name_container.data_envelopes) != 1:
                return None
            else:
                index_props: list[str] = collection_name_container.data_envelopes[0][envelope_payload_][index_props_]
                collection_name = collection_name_container.data_envelopes[0][ReservedPropName.collection_name.name]

                arg_name_to_prop_name_map: list[dict] = []
                for index_prop in index_props:
                    # For metadata: `arg_name` is the same as prop name:
                    arg_name_to_prop_name_map.append({index_prop: index_prop})

                # Construct `search_control` dynamically:
                data_envelope_search_control_dict = populate_search_control(
                    collection_name,
                    {},
                    arg_name_to_prop_name_map,
                )
                data_envelope_search_control_obj = search_control_desc.dict_schema.load(
                    data_envelope_search_control_dict,
                )
                return data_envelope_search_control_obj

        else:
            return None



from marshmallow import (
    fields,
    RAISE,
)

from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_root.schema_config.ConnectionConfigSchema import connection_config_desc
from argrelay_schema_config_client.runtime_data_client_app.ClientConfig import ClientConfig

__comment___ = "__comment__"
redundant_servers_ = "redundant_servers"
use_local_requests_ = "use_local_requests"
optimize_completion_request_ = "optimize_completion_request"
show_pending_spinner_ = "show_pending_spinner"
spinless_sleep_sec_ = "spinless_sleep_sec"


# NOTE: Client does not use `Schema` to load config in prod code
#       (only in tests due to heavy import caused by `Schema`).
#       Therefore, validation and applying defaults is not done by this class.
#       Duplicate the same requirements (for validation and defaults) in client prod code.
class ClientConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = ClientConfig

    # Allow this field in JSON (otherwise schema validation fails):
    __comment__ = fields.String(
        required = False,
        load_default = "",
    )

    # Serve requests from local data or send to server
    # (used in test only - see `ClientLocal` and FS_66_17_43_42 test infra):
    use_local_requests = fields.Boolean(
        required = False,
        load_default = False,
    )

    # Use one of these (default = True):
    # *   if True: ClientCommandRemoteWorkerTextProposeArgValuesOptimized
    # *   if False: ClientCommandRemoteWorkerJson
    optimize_completion_request = fields.Boolean(
        required = False,
        load_default = True,
    )

    redundant_servers = fields.List(
        fields.Nested(connection_config_desc.dict_schema),
        required = True,
    )

    # Enables spinner for FS_14_59_14_06: pending requests.
    show_pending_spinner = fields.Boolean(
        required = False,
        load_default = False,
    )

    spinless_sleep_sec = fields.Number(
        required = False,
        # Noticeable threshold is around 200 ms, but default to spin immediately:
        load_default = 0.0,
    )


client_config_desc = TypeDesc(
    dict_schema = ClientConfigSchema(),
    ref_name = ClientConfigSchema.__name__,
    dict_example = {
        redundant_servers_: [
            connection_config_desc.dict_example,
        ]
    },
    default_file_path = "argrelay_client.json",
)



from __future__ import annotations

from argrelay_api_plugin_server_abstract.delegator_utils import redirect_to_error
from argrelay_api_plugin_server_abstract.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
)
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaCustomDataDelegatorError import (
    error_code_,
    error_delegator_stub_custom_data_example,
    error_message_,
)
from argrelay_lib_server_plugin_demo.demo_service.DelegatorServiceHostBase import (
    DelegatorServiceHostBase,
    get_host_search_control,
)
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)

func_id_desc_host_ = "func_id_desc_host"


class DelegatorServiceHostDesc(DelegatorServiceHostBase):

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        host_search_control = get_host_search_control()

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: func_id_desc_host_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
                ReservedPropName.help_hint.name: "Describe remote host",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_desc_host_,
            },
        ]
        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == func_id_desc_host_

        # Actual implementation is not defined for demo:
        return redirect_to_error(
            interp_ctx,
            local_server.plugin_config,
            error_delegator_stub_custom_data_example[error_message_],
            error_delegator_stub_custom_data_example[error_code_],
        )



from enum import (
    auto,
    Enum,
)


class CompositeInfoType(Enum):
    """
    Specifies info type extract-able from FS_33_76_82_84 the composite forest.
    """

    tree_abs_path_to_interp_id = auto()
    """
    Config for `tree_abs_path_to_interp_id` is used in delegator instances with `single_func_id` for special funcs.

    The `func_tree_node` is configured single `func_id` in the `sub_tree` (empty "" node name).

    The current path to `func_tree_node` is transformed back to the path of `zero_arg_node` leading to it.
    """

    # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor
    #       It is identical to `interp_tree`, but `interp_tree` does not include zero arg components.
    #       Consider making `interp_tree` global (with full paths including zero arg interp) -
    #       this way there will be no need for separate extraction of `zero_arg_interp_tree`.
    zero_arg_interp_tree = auto()
    """
    See FS_42_76_93_51 very first zero arg mapping interp.
    """

    jump_tree = auto()
    """
    See FS_91_88_07_23 jump tree.

    A tree of `CompositeNodeType.interp_tree_node` under `CompositeNodeType.zero_arg_node`
    with leaves set to absolute jump path within `composite_forest`.
    """

    interp_tree = auto()
    """
    See FS_01_89_09_24 interp tree.

    A tree of `CompositeNodeType.interp_tree_node` under `CompositeNodeType.zero_arg_node`
    with leaves set to `plugin-instance_id`-s derived from `AbstractInterpFactory`.
    """

    func_tree = auto()
    """
    See FS_26_43_73_72 func tree.

    A tree of `CompositeNodeType.func_tree_node` under `CompositeNodeType.interp_tree_node`
    with leaves set to `func_id`-s.
    """



from copy import deepcopy

from marshmallow import (
    fields,
    RAISE,
)

from argrelay_api_server_cli.schema_response.ArgValuesSchema import (
    arg_values_desc,
    ArgValuesSchema,
)
from argrelay_api_server_cli.schema_response.EnvelopeContainerSchema import envelope_container_desc
from argrelay_api_server_cli.schema_response.InterpResult import InterpResult
from argrelay_lib_root.enum_desc.SpecialChar import SpecialChar
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc

"""
Schema for the result of interpretation taken from :class:`InterpContext`
"""

all_tokens_ = "all_tokens"
excluded_tokens_ = "excluded_tokens"
consumed_token_buckets_ = "consumed_token_buckets"
envelope_containers_ = "envelope_containers"
tan_token_ipos_ = "tan_token_ipos"
tan_token_l_part_ = "tan_token_l_part"


class InterpResultSchema(ArgValuesSchema):
    """
    See also `InterpResult`.
    """

    class Meta:
        unknown = RAISE
        ordered = True

    model_class = InterpResult

    all_tokens = fields.List(
        fields.String(),
        required = True,
    )

    excluded_tokens = fields.List(
        fields.Integer(),
        required = True,
    )

    consumed_token_buckets = fields.List(
        fields.List(
            fields.Integer(),
        ),
        required = True,
    )

    envelope_containers = fields.List(
        fields.Nested(envelope_container_desc.dict_schema),
        required = True,
    )

    tan_token_ipos = fields.Integer()

    tan_token_l_part = fields.String()


_interp_result_example = deepcopy(arg_values_desc.dict_example)
_interp_result_example.update({
    all_tokens_: [
        "some_command",
        "unrecognized_token",
        "goto",
        "host",
        "prod",
        SpecialChar.TokenBucketDelimiter.value,
    ],
    excluded_tokens_: [
        5,
    ],
    consumed_token_buckets_: [
        [
            0,
            2,
            3,
            4,
        ],
        [],
    ],
    envelope_containers_: [
        envelope_container_desc.dict_example,
    ],
    tan_token_ipos_: 1,
    tan_token_l_part_: "unrecognized_",
})

interp_result_desc = TypeDesc(
    dict_schema = InterpResultSchema(),
    ref_name = InterpResultSchema.__name__,
    dict_example = _interp_result_example,
    default_file_path = "",
)



from __future__ import annotations

from marshmallow import (
    fields,
    RAISE,
)

from argrelay_api_server_cli.schema_response.EnvelopeContainerSchema import data_envelopes_
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_server.runtime_data_server_app.EnvelopeCollection import EnvelopeCollection
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import data_envelope_desc

collection_name_ = "collection_name"


class EnvelopeCollectionSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = EnvelopeCollection

    collection_name = fields.String(
        required = True,
    )

    data_envelopes = fields.List(
        fields.Nested(data_envelope_desc.dict_schema),
        required = False,
        load_default = [],
    )


envelope_collection_desc = TypeDesc(
    dict_schema = EnvelopeCollectionSchema(),
    ref_name = EnvelopeCollectionSchema.__name__,
    dict_example = {
        collection_name_: ReservedEnvelopeClass.class_function.name,
        data_envelopes_: [
            data_envelope_desc.dict_example,
        ],
    },
    default_file_path = "",
)



from __future__ import annotations

import time
import uuid
from copy import deepcopy
from typing import Union

from pymongo import MongoClient

from argrelay_api_plugin_server_abstract.AbstractLoader import AbstractLoader
from argrelay_api_plugin_server_abstract.AbstractPluginServer import (
    AbstractPluginServer,
    instantiate_server_plugin,
)
from argrelay_api_plugin_server_abstract.DelegatorAbstract import DelegatorAbstract
from argrelay_app_server.composite_forest.CompositeForestValidator import validate_composite_forest
from argrelay_app_server.composite_forest.DictTreeWalker import contains_whitespace
from argrelay_app_server.mongo_data import MongoClientWrapper
from argrelay_app_server.mongo_data.MongoServerWrapper import MongoServerWrapper
from argrelay_app_server.mongo_data.ProgressTracker import ProgressTracker
from argrelay_app_server.relay_server.HelpHintCache import HelpHintCache
from argrelay_app_server.relay_server.IndexModel import (
    index_props_,
    IndexModel,
)
from argrelay_app_server.relay_server.QueryEngine import QueryEngine
from argrelay_app_server.relay_server.UsageStatsStore import UsageStatsStore
from argrelay_app_server.runtime_context.SearchControl import SearchControl
from argrelay_lib_root.enum_desc.PluginSide import PluginSide
from argrelay_lib_root.enum_desc.PluginType import PluginType
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialChar import SpecialChar
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_lib_server_plugin_core.plugin_interp.AbstractInterpFactory import AbstractInterpFactory
from argrelay_schema_config_server.runtime_data_server_app.EnvelopeCollection import EnvelopeCollection
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig
from argrelay_schema_config_server.runtime_data_server_plugin.PluginConfig import PluginConfig
from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    envelope_id_,
    envelope_payload_,
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import search_control_list_
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import (
    arg_name_to_prop_name_map_,
    collection_name_,
    search_control_desc,
)
from argrelay_schema_config_server.schema_config_server_app.EnvelopeCollectionSchema import (
    envelope_collection_desc,
)


class LocalServer:
    """
    This is plain server functionality without API-wrapper to expose over the network (hence, local).

    The API-wrapper exposing `LocalServer` over the network is `CustomFlaskApp`.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_config: PluginConfig,
    ):
        self.server_instance_id = uuid.uuid4()
        self.server_config: ServerConfig = server_config
        self.plugin_config: PluginConfig = plugin_config
        self.mongo_server: MongoServerWrapper = MongoServerWrapper()
        self.mongo_client: MongoClient = MongoClientWrapper.get_mongo_client(self.server_config.mongo_config)
        self.usage_stats_store = UsageStatsStore()
        self.query_engine: QueryEngine = QueryEngine(
            self.server_config.query_cache_config,
            self.get_mongo_database(),
            self.server_config.mongo_config.distinct_values_query,
        )
        self.help_hint_cache: HelpHintCache = HelpHintCache(
            self.query_engine,
        )

        self.cleaned_mongo_collections: set[str] = set()
        """
        *   On one hand, Mongo collections may have state and need to be cleaned.
        *   On another hand, they can be repeatedly updated by more than one plugin (potentially adding more data).

        Server uses `cleaned_mongo_collections` to track what collection have been cleaned once and what have not.
        """

        # FS_45_08_22_15 index model:
        # This internal server state is a convenience cache for index model which is also stored in data backend.
        self.index_model_per_collection: dict[str, IndexModel] = {}

        # seconds since epoch:
        self.server_start_time: int = int(time.time())

        self.root_func_loader: Union[AbstractInterpFactory | None] = None

    def start_local_server(
        self,
    ):
        self._instantiate_plugins()
        self._validate_composite_forest()
        self._activate_plugins()
        self._start_mongo_server()
        self._store_mongo_data()
        self._populate_help_hint_cache()
        self._log_connection_url()

    def stop_local_server(
        self,
    ):
        self._stop_mongo_server()

    def get_mongo_database(
        self,
    ):
        return self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

    def get_query_engine(
        self,
    ):
        return self.query_engine

    def _validate_composite_forest(
        self,
    ):
        validate_composite_forest(
            self.server_config.server_plugin_control.composite_forest,
            self.server_config.plugin_instances,
        )

    def _instantiate_plugins(
        self,
    ):
        """
        Instantiates every plugin.
        """

        for plugin_instance_id in self.plugin_config.plugin_instance_id_activate_list:
            plugin_entry: PluginEntry = self.plugin_config.server_plugin_instances[plugin_instance_id]

            if not plugin_entry.plugin_enabled:
                continue

            if (
                plugin_entry.plugin_side != PluginSide.PluginServerSideOnly
                and
                plugin_entry.plugin_side != PluginSide.PluginAnySide
            ):
                continue

            plugin_instance: AbstractPluginServer = instantiate_server_plugin(
                self.server_config,
                plugin_instance_id,
                plugin_entry,
            )
            plugin_type = plugin_instance.get_plugin_type()

            assert plugin_type.plugin_side is plugin_entry.plugin_side

            self.server_config.plugin_instances[plugin_instance_id] = plugin_instance

            if plugin_type is PluginType.LoaderPlugin:
                plugin_instance: AbstractLoader
                # Store instance of `AbstractLoader` under specified id for future use:
                self.server_config.data_loaders[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.InterpFactoryPlugin:
                plugin_instance: AbstractInterpFactory
                # Store instance of `AbstractInterpFactory` under specified id for future use:
                self.server_config.interp_factories[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.DelegatorPlugin:
                plugin_instance: DelegatorAbstract
                # Store instance of `DelegatorAbstract` under specified id for future use:
                self.server_config.action_delegators[plugin_instance_id] = plugin_instance
                continue

            if plugin_type is PluginType.ConfiguratorPlugin:
                plugin_instance: DelegatorAbstract
                # Store instance of `ConfiguratorAbstract` under specified id for future use:
                self.server_config.server_configurators[plugin_instance_id] = plugin_instance
                continue

    def _activate_plugins(
        self,
    ):
        for plugin_instance_id in self.plugin_config.plugin_instance_id_activate_list:
            # Skip those which were not instantiated:
            if plugin_instance_id in self.server_config.plugin_instances:
                self.server_config.plugin_instances[plugin_instance_id].activate_plugin()

        interp_factory: AbstractInterpFactory
        for plugin_instance_id, interp_factory in self.server_config.interp_factories.items():
            if interp_factory.is_root_func_loader():
                # There can only be one `root_func_loader`:
                assert self.root_func_loader is None
                self.root_func_loader = interp_factory

    def _pre_validate_static_data_per_step(
        self,
        envelope_collections: list[EnvelopeCollection],
        progress_tracker: ProgressTracker,
    ):
        self._pre_validate_static_data_schema_per_step(
            envelope_collections,
        )
        self._pre_validate_data_envelope_string_index_prop_values_per_step(
            envelope_collections,
            progress_tracker,
        )

    # noinspection PyMethodMayBeStatic
    def _pre_validate_static_data_schema_per_step(
        self,
        envelope_collections: list[EnvelopeCollection],
    ):
        # Note that this is slow for large data sets:
        for envelope_collection_obj in envelope_collections:
            envelope_collection_dict = envelope_collection_desc.dict_schema.dump(envelope_collection_obj)
            envelope_collection_desc.validate_dict(envelope_collection_dict)

    def _pre_validate_data_envelope_string_index_prop_values_per_step(
        self,
        envelope_collections: list[EnvelopeCollection],
        progress_tracker: ProgressTracker,
    ):
        """
        This validation ensures there is no blank values (`None`, whitespace, etc.) in `index_prop`-s.

        TODO: TODO_39_25_11_76: `data_envelope`-s with missing props.

        See also FS_99_81_19_25 no space in options.

        See also `_post_validate_all_data_envelope_for_missing_index_prop_names`.
        """

        for envelope_collection in envelope_collections:
            collection_name = envelope_collection.collection_name
            progress_tracker.track_collection_size_increment(
                collection_name,
                len(envelope_collection.data_envelopes),
            )
            for data_envelope in envelope_collection.data_envelopes:
                index_props = self.index_model_per_collection[collection_name].index_props
                for index_prop in index_props:
                    if index_prop not in data_envelope:
                        # Let the `data_envelope` load without some `prop_name`-s from `index_prop`-s
                        # in pre-validation - if any of `data_envelope`-s have that `prop_name`, it will fail
                        # in post-validation (see where `_raise_validation_error_for_missing_prop_name` is used).
                        # For now:
                        #     It is allowed to have no `prop_name` for all `data_envelope`
                        #     within given `collection_name` until at least one has it.
                        # TODO: TODO_39_25_11_76: `data_envelope`-s with missing props:
                        #       Be careful with "until at least one has it" as this
                        #       delayed validation has to be explicitly/separately run
                        #       after data is loaded.
                        #       Consider refactoring data manipulation into safe manager.
                        continue

                    prop_value = data_envelope[index_prop]
                    self._validate_string_prop_value(
                        collection_name,
                        index_prop,
                        prop_value,
                    )

    def _validate_string_prop_value(
        self,
        collection_name,
        index_prop,
        prop_value,
    ):
        if isinstance(prop_value, str):
            if not prop_value or not prop_value.strip():
                raise ValueError(f"`{collection_name}.{index_prop}` [{prop_value}] has to be non-blank string")
            if contains_whitespace(prop_value):
                raise ValueError(f"`{collection_name}.{index_prop}` [{prop_value}] cannot contain whitespace")
        elif isinstance(prop_value, list):
            for prop_value_item in prop_value:
                self._validate_string_prop_value(
                    collection_name,
                    index_prop,
                    prop_value_item,
                )
        else:
            # FS_06_99_43_60: array `prop_value`:
            raise ValueError(f"`{collection_name}.{index_prop}` has to be a `list` of `str` or `str`")

    def _post_validate_stored_data(
        self,
        progress_tracker: ProgressTracker,
    ):
        self._post_validate_all_data_envelope_for_missing_index_prop_names(
            progress_tracker,
        )
        self._post_validate_search_props_and_index_props_reference_each_other(
            self._scan_for_all_search_props(),
            progress_tracker,
        )

    def _post_validate_search_props_and_index_props_reference_each_other(
        self,
        search_props_per_collection: dict[str, set[str]],
        progress_tracker: ProgressTracker,
    ):
        """
        See TODO_39_25_11_76 missing props based on `search_control`-s for all loaded functions.

        This function relies on `_post_validate_all_data_envelope_for_missing_index_prop_names`
        to guarantee that all listed `index_props` exists in every loaded `data_envelope`.
        On top of that fact, this function ensures that every `search_control` used
        specifies `search_props` which exists among `index_props`.
        """

        # TODO: Deduplicate with L <- R:
        # Compare L -> R:
        # Ensure that `index_model` references are used in  some`search_control`.
        for collection_name, index_model in self.index_model_per_collection.items():
            if collection_name not in search_props_per_collection:
                progress_tracker.track_unused_index_props_per_collection_unused_by_search_control(
                    collection_name,
                    # all are unused:
                    index_model.index_props,
                )
            else:
                search_props = search_props_per_collection[collection_name]
                unused_index_props = []
                for index_prop in index_model.index_props:
                    if index_prop not in search_props:
                        unused_index_props.append(index_prop)
                if len(unused_index_props) > 0:
                    progress_tracker.track_unused_index_props_per_collection_unused_by_search_control(
                        collection_name,
                        # unused are selected only:
                        unused_index_props,
                    )

        # TODO: Deduplicate with L -> R:
        # Compare L <- R:
        # Ensure that every `search_prop` is among the `index_prop`-s in `index_model`:
        for collection_name, search_props in search_props_per_collection.items():
            if collection_name not in self.index_model_per_collection:
                progress_tracker.track_dangling_search_props_per_collection_undefined_by_index_model(
                    collection_name,
                    # all are dangling:
                    search_props,
                )
            else:
                search_props = search_props_per_collection[collection_name]
                index_props = self.index_model_per_collection[collection_name].index_props
                dangling_search_props = []
                for search_prop in search_props:
                    if search_prop not in index_props:
                        dangling_search_props.append(search_prop)
                if len(dangling_search_props) > 0:
                    progress_tracker.track_dangling_search_props_per_collection_undefined_by_index_model(
                        collection_name,
                        # dangling are selected only:
                        dangling_search_props,
                    )

    def _scan_for_all_search_props(
        self,
    ) -> dict[str, set[str]]:
        """
        Build search `prop_name`-s per `collection_name`
        by scanning `search_control`-s for `prop_name`-s.
        """

        search_props_per_collection: dict[str, set[str]] = {}

        plugin_search_controls: list[SearchControl] = []
        for plugin_instance in self.server_config.plugin_instances.values():
            plugin_search_controls.extend(plugin_instance.provide_plugin_search_control())

        # Scan `search_control`-s of all plugins:
        for search_control in plugin_search_controls:
            self._populate_search_props_per_collection(
                search_control_desc.dict_from_input_obj(search_control),
                search_props_per_collection,
            )

        # Scan `search_control`-s of all funcs:
        for func_envelope in self.query_engine.get_data_envelopes_cursor(
            ReservedEnvelopeClass.class_function.name,
            {
                f"{ReservedPropName.envelope_class.name}": f"{ReservedEnvelopeClass.class_function.name}",
            },
        ):
            for search_control in func_envelope[instance_data_][search_control_list_]:
                self._populate_search_props_per_collection(
                    search_control,
                    search_props_per_collection,
                )

        return search_props_per_collection

    # noinspection PyMethodMayBeStatic
    def _populate_search_props_per_collection(
        self,
        search_control: dict,
        search_props_per_collection,
    ):
        search_collection_name = search_control[collection_name_]

        search_props = search_props_per_collection.setdefault(
            search_collection_name,
            set(),
        )

        for arg_name_to_prop_name_entry in search_control[arg_name_to_prop_name_map_]:
            assert type(arg_name_to_prop_name_entry) is dict
            assert len(arg_name_to_prop_name_entry) == 1
            search_prop_name = next(iter(arg_name_to_prop_name_entry.values()))
            search_props.add(search_prop_name)

    def _post_validate_all_data_envelope_for_missing_index_prop_names(
        self,
        progress_tracker: ProgressTracker,
    ):
        """
        See TODO_39_25_11_76 missing props based on `index_prop`-s per `envelope_collection`.

        See also:
        *   `_pre_validate_data_envelope_string_index_prop_values_per_step`
        *   `_post_validate_search_props_and_index_props_reference_each_other`
        """

        progress_tracker.track_total_validation_start()

        # Verify that all prop names per `collection_name`
        # exists in all corresponding `data_envelope`-s:
        for collection_name in self.index_model_per_collection.keys():
            progress_tracker.track_collection_validation_start(collection_name)
            self.post_validate_collection_for_missing_index_prop_names(
                collection_name,
                progress_tracker,
            )
            progress_tracker.track_collection_validation_stop(collection_name)

        progress_tracker.track_total_validation_stop()

    def post_validate_collection_for_missing_index_prop_names(
        self,
        collection_name,
        progress_tracker,
    ):
        index_model = self.index_model_per_collection[collection_name]
        prev_found_missing_prop_names: dict[str, dict] = {}
        prev_found_existing_prop_names: dict[str, dict] = {}
        for data_envelope in self.query_engine.get_data_envelopes_cursor(
            collection_name,
            {},
        ):
            # TODO: Even if we track and validate `data_envelope` count on server start,
            #       we do not validate (only track increments) for `DelegatorDataBackendSet`
            #       for now - to do this, we need know how many `data_envelope`-s each
            #       set operation replaces.
            progress_tracker.track_collection_validation_increment(collection_name)

            for prop_name in index_model.index_props:
                if prop_name not in data_envelope:

                    if prop_name not in prev_found_missing_prop_names:
                        prev_found_missing_prop_names[prop_name] = deepcopy(data_envelope)

                    if prop_name in prev_found_existing_prop_names:
                        self._raise_validation_error_for_missing_prop_name(
                            True,
                            collection_name,
                            prop_name,
                            data_envelope,
                            prev_found_existing_prop_names[prop_name],
                        )
                else:

                    if prop_name not in prev_found_existing_prop_names:
                        prev_found_existing_prop_names[prop_name] = deepcopy(data_envelope)

                    if prop_name in prev_found_missing_prop_names:
                        self._raise_validation_error_for_missing_prop_name(
                            False,
                            collection_name,
                            prop_name,
                            data_envelope,
                            prev_found_missing_prop_names[prop_name],
                        )
                    prop_value = data_envelope[prop_name]
                    self._validate_string_prop_value(
                        collection_name,
                        prop_name,
                        prop_value,
                    )

    # noinspection PyMethodMayBeStatic
    def _raise_validation_error_for_missing_prop_name(
        self,
        is_missing_otherwise_existing_prop_name: bool,
        collection_name: str,
        prop_name: str,
        curr_sample: dict,
        prev_sample: dict,
    ):
        """
        Raise informative exception for cases of TODO_39_25_11_76: `data_envelope`-s with missing props.
        """

        # It is allowed to have no `prop_name` for all `data_envelope` until at least one has it.
        if is_missing_otherwise_existing_prop_name:
            raise ValueError(
                f"`data_envelope` of `collection_name` [{collection_name}] does not have `prop_name` [{prop_name}] while another one had:\ncurr_sample:\n{curr_sample}\nprev_sample:\n{prev_sample}"
            )
        else:
            raise ValueError(
                f"`data_envelope` of `collection_name` [{collection_name}] has `prop_name` [{prop_name}] while another one did not have:\ncurr_sample:\n{curr_sample}\nprev_sample:\n{prev_sample}"
            )

    def _start_mongo_server(
        self,
    ):
        self.mongo_server.start_mongo_server(self.server_config.mongo_config)

    def _stop_mongo_server(
        self,
    ):
        self.mongo_server.stop_mongo_server()

    # noinspection PyMethodMayBeStatic
    def _populate_func_missing_dynamic_props_(
        self,
        func_envelope_collection: EnvelopeCollection,
        func_index_props: list[str],
    ):
        """
        TODO: TODO_39_25_11_76 missing props: populate missing props supported by funcs.

        This function only targets `ReservedEnvelopeClass.class_function` (because loading funcs is special).
        There is no decision (yet) to populate missing `prop_value`-s for any `data_envelope`.
        Instead, there is a validation to prevent missing `prop_name`-s for other `data_envelope`-s - if it fails,
        the corresponding loader has to be fixed to provide seme set of `prop_name`-s for all `data_envelope`.
        """

        # Populate missing `prop_name`-s with special value:
        for data_envelope in func_envelope_collection.data_envelopes:
            envelope_class = data_envelope[ReservedPropName.envelope_class.name]
            if envelope_class == ReservedEnvelopeClass.class_function.name:
                for prop_name in func_index_props:
                    if prop_name not in data_envelope:
                        data_envelope[prop_name] = SpecialChar.NoPropValue.value

    def _store_mongo_data(
        self,
    ):

        progress_tracker = ProgressTracker()

        # At this moment, funcs have already been loaded on `AbstractPlugin.activate_plugin`.
        func_envelope_collection = EnvelopeCollection(
            collection_name = ReservedEnvelopeClass.class_function.name,
            data_envelopes = self.root_func_loader.get_func_data_envelopes(),
        )
        self._define_func_index_model(
            func_envelope_collection,
        )
        # Initial step: store funcs and any data from config:
        self.store_mongo_data_step(
            # Single `envelope_collection` with `func_data_envelopes`:
            [func_envelope_collection],
            "config_data",
            progress_tracker,
        )

        for plugin_instance_id in self.plugin_config.plugin_instance_id_activate_list:
            if plugin_instance_id in self.server_config.data_loaders:
                plugin_instance = self.server_config.data_loaders[plugin_instance_id]

                # Use loader to define required FS_45_08_22_15 index models.
                index_models: list[IndexModel] = plugin_instance.list_index_models()

                for index_model in index_models:
                    self._define_index_model_step(
                        index_model,
                    )

                # Use loader to update data:
                envelope_collections: list[EnvelopeCollection] = plugin_instance.load_envelope_collections(
                    self.query_engine,
                )

                self.store_mongo_data_step(
                    envelope_collections,
                    plugin_instance_id,
                    progress_tracker,
                )

        self._store_meta_data(
            progress_tracker,
        )

        self._post_validate_stored_data(progress_tracker)

    def _store_meta_data(
        self,
        progress_tracker,
    ):
        """
        This method stores FS_45_08_22_15 index model into data backend.

        The index model (metadata) is used for FS_74_69_61_79 get set data envelope.
        """

        # Add generated `envelope_collection` (itself) for completeness:
        self._define_index_model_step(
            IndexModel(
                collection_name = ReservedEnvelopeClass.class_collection.name,
                index_props = [
                    # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
                    ReservedPropName.envelope_class.name,
                    ReservedPropName.collection_name.name,
                ],
            ),
        )

        envelope_collection = EnvelopeCollection(
            collection_name = ReservedEnvelopeClass.class_collection.name,
            data_envelopes = []
        )
        for collection_name, index_model in self.index_model_per_collection.items():
            envelope_collection.data_envelopes.append({
                envelope_id_: f"{collection_name}",
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_collection.name,
                # TODO: May add loader plugin instance name as metadata:
                ReservedPropName.collection_name.name: collection_name,
                envelope_payload_: {
                    index_props_: list(index_model.index_props),
                },
            })

        self.store_mongo_data_step(
            [envelope_collection],
            self.__class__.__name__,
            progress_tracker,
        )

    def _define_func_index_model(
        self,
        func_envelope_collection: EnvelopeCollection,
    ):
        """
        Implements FS_45_08_22_15 index model.

        Unlike most of other data, func data is not loaded via loaders and its index model is defined here.
        """

        func_index_props: list[str] = self.root_func_loader.get_func_dynamic_index_props()

        self._populate_func_missing_dynamic_props_(
            func_envelope_collection,
            func_index_props,
        )

        # These added after `_populate_func_missing_dynamic_props_` to ensure that they are populated explicitly.
        # Extend `index_props` for funcs:
        func_index_props.extend([
            # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
            ReservedPropName.envelope_class.name,
            ReservedPropName.func_id.name,
            ReservedPropName.func_state.name,
        ])

        func_index_model: IndexModel = IndexModel(
            collection_name = ReservedEnvelopeClass.class_function.name,
            index_props = func_index_props,
        )

        self._define_index_model_step(func_index_model)

    def _define_index_model_step(
        self,
        given_index_model: IndexModel,
    ):
        """
        Implements internal API for FS_45_08_22_15 index model.

        If there are already existing list of `index_props` per `class_name` per `collection_name`,
        it is possible to call it again, but `index_props` must be exactly the same.
        This is to allow multiple loaders define data they work with,
        but that data definition must match across all loaders working with this type.

        TODO: Limitation: this is append only at the moment
              (you cannot remove added index model - you can only extend or restart server to start over).

        TODO: Expose this via REST API.
        """

        known_index_models: dict[str, IndexModel] = self.index_model_per_collection

        # Ensure `given_index_model` does not redefine `known_index_model` anyhow:
        if given_index_model.collection_name not in known_index_models:
            known_index_models[given_index_model.collection_name] = given_index_model
        else:
            known_index_model = known_index_models[given_index_model.collection_name]
            # Compare L to R:
            for known_index_prop in known_index_model.index_props:
                assert known_index_prop in given_index_model.index_props
            # Compare R to L:
            for given_index_prop in given_index_model.index_props:
                assert given_index_prop in known_index_model.index_props

    def delete_data_envelopes(
        self,
        collection_name: str,
        query_dict: dict,
    ):
        mongo_db = self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]
        MongoClientWrapper.delete_data_envelopes(
            mongo_db,
            collection_name,
            query_dict,
        )

    def store_mongo_data_step(
        self,
        envelope_collections: list[EnvelopeCollection],
        step_name: str,
        progress_tracker: ProgressTracker,
    ):
        mongo_db = self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

        eprint(f"step name: {step_name}:")

        self._pre_validate_static_data_per_step(
            envelope_collections,
            progress_tracker,
        )

        MongoClientWrapper.store_envelopes(
            mongo_db,
            self.cleaned_mongo_collections,
            envelope_collections,
            progress_tracker,
        )

        self._create_mongo_index_step(
            envelope_collections,
        )

    def invalidate_cache_for_collection(
        self,
        collection_name: str,
    ):
        self.query_engine.invalidate_cache_for_collection(collection_name)

    def _create_mongo_index_step(
        self,
        envelope_collections: list[EnvelopeCollection],
    ):
        mongo_db = self.mongo_client[self.server_config.mongo_config.mongo_server.database_name]

        for envelope_collection in envelope_collections:
            collection_name = envelope_collection.collection_name

            index_props: list[str] = self.index_model_per_collection[collection_name].index_props

            MongoClientWrapper.create_index(
                mongo_db,
                collection_name,
                index_props,
            )

    def _populate_help_hint_cache(
        self,
    ):
        self.help_hint_cache.populate_cache()

    def _log_connection_url(
        self,
    ):
        host_name = self.server_config.connection_config.server_host_name
        port_number = self.server_config.connection_config.server_port_number
        eprint(f"http://{host_name}:{port_number}")



from __future__ import annotations

from typing import (
    Type,
    Union,
)

from argrelay_api_plugin_server_abstract.DelegatorAbstract import DelegatorAbstract
from argrelay_api_server_cli.schema_response.AssignedValue import AssignedValue
from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_server.runtime_context.EnvelopeContainer import EnvelopeContainer
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_lib_root.enum_desc.ValueSource import ValueSource
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_test_infra.test_infra.BaseTestClass import BaseTestClass
from argrelay_test_infra.test_infra.EnvMockBuilder import EnvMockBuilder


class InOutTestClass(BaseTestClass):
    # TODO: TODO_32_99_70_35: Make generic validator be able to verify payload (not only `interp_ctx` passed from local client) - JSONPath?

    # TODO: Wrap input into Dataclass which in turn can be created via builder (to have ability pre-build defaults for a set of tests).
    def verify_response_data(
        self,
        call_ctx: CallContext,
        actual_suggestions: list[str],
        container_ipos_to_expected_assignments: Union[dict[int, dict[str, AssignedValue]], None],
        container_ipos_to_options_hidden_by_default_value: Union[dict[int, dict[str, list[str]]], None],
        delegator_class: Union[Type[DelegatorAbstract], None],
        envelope_ipos_to_prop_values: Union[dict[int, dict[str, str]], None],
        expected_suggestions: Union[list[str], None],
        envelope_containers: list[EnvelopeContainer],
        expected_container_ipos_to_used_token_bucket: Union[dict[int, Union[int, None]], None],
    ):
        try:

            if expected_suggestions is not None:
                self.assertEqual(expected_suggestions, actual_suggestions)

            if container_ipos_to_expected_assignments is not None:
                self.verify_envelope_containers(
                    envelope_containers,
                    container_ipos_to_expected_assignments,
                )

            if container_ipos_to_options_hidden_by_default_value is not None:
                self.verify_options_hidden_by_default_value(
                    envelope_containers,
                    container_ipos_to_options_hidden_by_default_value,
                )

            if (
                container_ipos_to_expected_assignments is not None
                and
                container_ipos_to_options_hidden_by_default_value is not None
            ):
                self.verify_expected_assignments_with_option_type_hidden_by_default(
                    container_ipos_to_expected_assignments,
                    container_ipos_to_options_hidden_by_default_value
                )

            if delegator_class is not None:
                self.assertEqual(
                    EnvMockBuilder.invocation_input.delegator_plugin_entry.plugin_module_name,
                    delegator_class.__module__,
                )
                self.assertEqual(
                    EnvMockBuilder.invocation_input.delegator_plugin_entry.plugin_class_name,
                    delegator_class.__name__,
                )

            if envelope_ipos_to_prop_values is not None:
                self.verify_data_envelopes(
                    EnvMockBuilder.invocation_input.get_data_envelopes(),
                    envelope_ipos_to_prop_values,
                )

            if expected_container_ipos_to_used_token_bucket is not None:
                # TODO: TODO_32_99_70_35: candidate for generic library of JSONPath verifiers:
                for envelope_container_ipos, envelope_container in enumerate(envelope_containers):
                    expected_used_token_bucket = expected_container_ipos_to_used_token_bucket[envelope_container_ipos]
                    self.assertEqual(
                        expected_used_token_bucket,
                        envelope_container.used_token_bucket,
                    )

        except:

            # Check if there are some obvious test input mistakes:

            if call_ctx.server_action is ServerAction.ProposeArgValues:
                self.assertIsNone(delegator_class)
                self.assertIsNone(envelope_ipos_to_prop_values)
            elif call_ctx.server_action is ServerAction.DescribeLineArgs:
                # TODO: TODO_42_81_01_90: assert data instead of print out: there will be no delegator_class, but a mock to intercept response for enum query:
                self.assertIsNone(delegator_class)
                # TODO: TODO_42_81_01_90: assert data instead of print out: the data should be available:
                self.assertIsNone(envelope_ipos_to_prop_values)
            elif call_ctx.server_action is ServerAction.RelayLineArgs:
                pass
            else:
                raise RuntimeError

            # Rethrow previous error:
            raise

    def verify_envelope_containers(
        self,
        envelope_containers,
        container_ipos_to_expected_assignments: Union[dict[int, dict[str, AssignedValue]], None],
    ):
        for container_ipos, expected_assignments in container_ipos_to_expected_assignments.items():
            try:
                if expected_assignments is None:
                    self.assertFalse(0 <= container_ipos < len(envelope_containers))
                else:
                    self.assertTrue(0 <= container_ipos < len(envelope_containers))
                    for prop_name, assigned_value in expected_assignments.items():
                        try:
                            if assigned_value is None:
                                self.assertTrue(
                                    prop_name not in
                                    envelope_containers
                                    [container_ipos].assigned_prop_name_to_prop_value
                                )
                            elif isinstance(assigned_value, list):
                                # A bit of hack: if `list`, then check if it matches `remaining_prop_name_to_prop_value`:
                                self.assertEqual(
                                    assigned_value,
                                    envelope_containers
                                    [container_ipos].remaining_prop_name_to_prop_value
                                    [prop_name]
                                )
                            else:
                                self.assertEqual(
                                    assigned_value,
                                    envelope_containers
                                    [container_ipos].assigned_prop_name_to_prop_value
                                    [prop_name],
                                )
                        except:
                            eprint(
                                f"container_ipos:{container_ipos} prop_name:{prop_name} assigned_value:{assigned_value}",
                            )
                            raise
            except:
                print(f"envelope_containers:\n{envelope_containers}")
                raise

    def verify_options_hidden_by_default_value(
        self,
        envelope_containers: list[EnvelopeContainer],
        container_ipos_to_options_hidden_by_default_value: Union[dict[int, dict[str, list[str]]], None],
    ):
        """
        Make sure that specified expected `options_hidden_by_default_value` list
        (per `envelope_container` ipos, per `prop_name`) match what is populated into
        `EnvelopeContainer.filled_prop_values_hidden_by_defaults`.
        """
        for container_ipos, options_hidden_by_default_value_per_prop_name in container_ipos_to_options_hidden_by_default_value.items():
            try:
                if options_hidden_by_default_value_per_prop_name is None:
                    self.assertFalse(0 <= container_ipos < len(envelope_containers))
                else:
                    self.assertTrue(0 <= container_ipos < len(envelope_containers))
                    for prop_name, options_hidden_by_default_value in options_hidden_by_default_value_per_prop_name.items():
                        try:
                            if options_hidden_by_default_value is None:
                                self.assertTrue(
                                    prop_name not in
                                    envelope_containers
                                    [container_ipos].filled_prop_values_hidden_by_defaults
                                )
                            else:
                                self.assertEqual(
                                    options_hidden_by_default_value,
                                    envelope_containers
                                    [container_ipos].filled_prop_values_hidden_by_defaults
                                    [prop_name],
                                )
                        except:
                            eprint(
                                f"container_ipos:{container_ipos} prop_name:{prop_name} options_hidden_by_default_value:{options_hidden_by_default_value}",
                            )
                            raise
            except:
                print(f"envelope_containers:\n{envelope_containers}")
                raise

    def verify_expected_assignments_with_option_type_hidden_by_default(
        self,
        container_ipos_to_expected_assignments: dict[int, dict[str, AssignedValue]],
        container_ipos_to_options_hidden_by_default_value: dict[int, dict[str, list[str]]],
    ):
        """
        Make sure that, for given `container_ipos`, if there are both:
        *   options hidden by default
        *   assigned value
        then, the assigned value has `ValueSource.default_value`.
        """
        for container_ipos, expected_assignments in container_ipos_to_expected_assignments.items():
            try:
                if expected_assignments is not None:
                    if container_ipos in container_ipos_to_options_hidden_by_default_value:
                        for prop_name, assigned_value in expected_assignments.items():
                            options_hidden_by_default_value_per_prop_name = (
                                container_ipos_to_options_hidden_by_default_value[
                                    container_ipos
                                ]
                            )
                            if prop_name in options_hidden_by_default_value_per_prop_name:
                                if options_hidden_by_default_value_per_prop_name[prop_name] is None:
                                    self.assertNotEqual(
                                        assigned_value.value_source,
                                        ValueSource.default_value,
                                    )
                                else:
                                    self.assertEqual(
                                        assigned_value.value_source,
                                        ValueSource.default_value,
                                    )
            except:
                print(
                    f"container_ipos_to_expected_assignments:\n{container_ipos_to_expected_assignments}"
                )
                print(
                    f"container_ipos_to_options_hidden_by_default_value:\n{container_ipos_to_options_hidden_by_default_value}"
                )
                raise

    def verify_data_envelopes(
        self,
        data_envelopes,
        envelope_ipos_to_prop_values,
    ):
        for envelope_ipos, prop_values in envelope_ipos_to_prop_values.items():
            try:
                if prop_values is None:
                    self.assertFalse(
                        (0 <= envelope_ipos < len(data_envelopes))
                        and
                        # TODO: Maybe, for clarity, avoid adding such None data envelope?
                        # Containers with non-found `data_envelope`-s still contribute an entry to this list:
                        (data_envelopes[envelope_ipos] is not None)
                    )
                else:
                    self.assertTrue(
                        (0 <= envelope_ipos < len(data_envelopes))
                    )
                    for prop_name, prop_value in prop_values.items():
                        try:
                            if prop_value is None:
                                self.assertTrue(
                                    prop_name not in
                                    data_envelopes[envelope_ipos]
                                )
                            else:
                                self.assertEqual(
                                    prop_value,
                                    data_envelopes[envelope_ipos][prop_name],
                                )
                        except:
                            eprint(
                                f"envelope_ipos:{envelope_ipos} prop_name:{prop_name} prop_value:{prop_value}",
                            )
                            raise

            except:
                print(f"data_envelopes:\n{data_envelopes}")
                raise



from __future__ import annotations

from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import (
    function_container_ipos_,
    InterpContext,
)
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
)

repo_container_ipos_ = 1

repo_root_abs_path_: str = "repo_root_abs_path"


class DelegatorGitRepoBase(DelegatorSingleFuncAbstract):
    """
    Implements base functionality for FS_67_16_61_97 git_plugin.
    """

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        # The first envelope (`DataEnvelopeSchema`) is assumed to be of
        # `ReservedEnvelopeClass.class_function` with `FunctionEnvelopeInstanceDataSchema` for its `instance_data`:
        function_container = interp_ctx.envelope_containers[function_container_ipos_]
        delegator_plugin_instance_id = (
            function_container.data_envelopes[0][instance_data_]
            [delegator_plugin_instance_id_]
        )
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input



from __future__ import annotations

from argrelay_api_plugin_server_abstract.delegator_utils import set_default_to
from argrelay_api_plugin_server_abstract.DelegatorAbstract import get_func_id_from_interp_ctx
from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import (
    function_container_ipos_,
    InterpContext,
)
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_lib_server_plugin_core.plugin_delegator.FuncConfigSchema import (
    fill_control_list_,
    func_envelope_,
)
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaConfigDelegatorConfigBase import (
    SchemaConfigDelegatorConfigBase,
)
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaConfigDelegatorConfigOnly import func_configs_
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import ServerConfig
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    envelope_payload_,
    instance_data_,
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
)

object_container_ipos_ = 1


class DelegatorConfigBase(DelegatorSingleFuncAbstract):
    """
    Part of FS_49_96_50_77 config_only_delegator implementation.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
        delegator_config_desc: TypeDesc,
    ):
        assert issubclass(delegator_config_desc.dict_schema.__class__, SchemaConfigDelegatorConfigBase)
        self.delegator_config_desc = delegator_config_desc

        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        """
        Populate (if missing) or assert (if present) func-related fields.
        """

        modified_plugin_config_dict = self.delegator_config_desc.dict_from_input_dict(plugin_config_dict)

        func_configs: dict = modified_plugin_config_dict[func_configs_]
        for func_id, func_config in func_configs.items():

            func_envelope = func_config[func_envelope_]

            if ReservedPropName.envelope_class.name in func_envelope:
                assert func_envelope[ReservedPropName.envelope_class.name] == ReservedEnvelopeClass.class_function.name
            else:
                func_envelope[ReservedPropName.envelope_class.name] = ReservedEnvelopeClass.class_function.name

            if ReservedPropName.func_id.name in func_envelope:
                assert func_envelope[ReservedPropName.func_id.name] == func_id
            else:
                func_envelope[ReservedPropName.func_id.name] = func_id

            instance_data = func_envelope[instance_data_]

            if func_id_ in instance_data:
                assert instance_data[func_id_] == func_id
            else:
                instance_data[func_id_] = func_id

            if delegator_plugin_instance_id_ in instance_data:
                assert instance_data[delegator_plugin_instance_id_] == self.plugin_instance_id
            else:
                instance_data[delegator_plugin_instance_id_] = self.plugin_instance_id

            if envelope_payload_ in func_envelope:
                envelope_payload = func_envelope[envelope_payload_]
            else:
                envelope_payload = {}
                func_envelope[envelope_payload_] = envelope_payload

        return modified_plugin_config_dict

    def validate_config(
        self,
    ) -> None:
        # Re-validate schema because `load_config` applies changes to config:
        self.delegator_config_desc.validate_dict(self.plugin_config_dict)

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        return [
            func_config[func_envelope_]
            for func_config in self.plugin_config_dict[func_configs_].values()
        ]

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def run_fill_control(
        self,
        interp_ctx: InterpContext,
    ) -> bool:
        """
        FS_49_96_50_77: (part of implementation for config-only delegator) populates defaults

        TODO: TODO_54_68_18_12: Support defaults for config-only delegator
        """
        any_assignment = False
        if interp_ctx.curr_container_ipos > function_container_ipos_:
            func_id = get_func_id_from_interp_ctx(interp_ctx)
            func_config = self.plugin_config_dict[func_configs_][func_id]
            if fill_control_list_ in func_config:
                fill_control_list = func_config[fill_control_list_]
                if interp_ctx.curr_container_ipos < len(fill_control_list):

                    # Define vars in current context to evaluate `prop_value_template` below:
                    envelope_containers = interp_ctx.envelope_containers

                    fill_control: dict = fill_control_list[interp_ctx.curr_container_ipos]
                    for prop_name, prop_value_template in fill_control.items():
                        if prop_value_template is not None:
                            # The input is trusted (from config), right? Then:
                            # https://stackoverflow.com/a/54071505/441652
                            prop_value = eval(f'f"""{prop_value_template}"""')

                            any_assignment = (
                                set_default_to(prop_name, prop_value, interp_ctx.curr_container)
                                or
                                any_assignment
                            )

        return any_assignment

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:

        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        # The first envelope (`DataEnvelopeSchema`) is assumed to be of
        # `ReservedEnvelopeClass.class_function` with `FunctionEnvelopeInstanceDataSchema` for its `instance_data`:
        function_container = interp_ctx.envelope_containers[function_container_ipos_]
        delegator_plugin_instance_id = (
            function_container.data_envelopes[0][instance_data_]
            [delegator_plugin_instance_id_]
        )
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        raise NotImplementedError("Extend: not implemented intentionally")



from __future__ import annotations

import os
import subprocess
from typing import Union

from icecream import ic

from argrelay_app_client.client_spec.ShellContext import (
    COMP_KEY_env_var,
    COMP_LINE_env_var,
    COMP_POINT_env_var,
    COMP_TYPE_env_var,
    UNKNOWN_COMP_KEY,
)
from argrelay_lib_root.enum_desc.CompType import CompType
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_test_infra.test_infra import parse_line_and_cpos
from argrelay_test_infra.test_infra.ClientServerTestClass import ClientServerTestClass
from argrelay_test_infra.test_infra.EnvMockBuilder import (
    EmptyEnvMockBuilder,
    EnvMockBuilder,
)


class End2EndTestClass(ClientServerTestClass):
    """
    Supports FS_66_17_43_42 test_infra / special test mode #5.

    In addition to starting server via its generated file in `@/bin/` dir (what `ClientServerTestClass` does),
    this tests also runs client via the generated file.

    Effectively, this runs both client and server outside the OS process responsible for running this test
    making all assertions via exit codes, stdout, stderr - what OS can provide as output of the OS process.

    It is probably "the fattest" test possible with end-to-end coverage while still using Python.
    """

    bound_command_env_var_name = "ARGRELAY_CLIENT_COMMAND"
    default_bound_command = os.environ.get(bound_command_env_var_name, "some_command")

    def env_vars(
        self,
        test_line: str,
        comp_type: CompType,
    ):
        (command_line, cursor_cpos) = parse_line_and_cpos(test_line)
        env_vars = os.environ.copy()
        env_vars[COMP_LINE_env_var] = command_line
        env_vars[COMP_POINT_env_var] = str(cursor_cpos)
        env_vars[COMP_TYPE_env_var] = str(comp_type.value)
        env_vars[COMP_KEY_env_var] = UNKNOWN_COMP_KEY
        return env_vars

    def run_client_with_cli_args(
        self,
        command_line_args,
        expected_exit_code,
    ):
        """
        See `CallConv.CliArgsConv`.
        """
        client_proc = subprocess.run(
            args = command_line_args,
            capture_output = True,
        )
        if expected_exit_code is not None:
            self.assertEqual(
                expected_exit_code,
                client_proc.returncode,
            )
        return client_proc

    def run_client_with_env_vars(
        self,
        command_name,
        env_vars,
        expected_exit_code,
    ):
        """
        See `CallConv.EnvVarsConv`.
        """
        client_proc = subprocess.run(
            args = [
                command_name,
            ],
            env = env_vars,
            capture_output = True,
        )
        if expected_exit_code is not None:
            self.assertEqual(
                expected_exit_code,
                client_proc.returncode,
            )
        return client_proc

    # TODO: allow have something on STDERR (even if it is wrong, we want this option to register known bugs to be fixed)
    def assert_no_stderr(
        self,
        stderr_str,
        server_action: ServerAction,
    ):
        self.assertEqual(
            "",
            stderr_str,
            f"{server_action} should have nothing on STDERR."
        )

    def assert_ProposeArgValues(
        self,
        command_name,
        test_line,
        comp_type: CompType,
        expected_stdout_str: Union[str, None],
        expected_exit_code = 0,
        env_mock_builder: EnvMockBuilder = None,
    ):
        if env_mock_builder is None:
            env_mock_builder = (
                EmptyEnvMockBuilder()
                .set_client_config_dict()
                .set_generate_client_config_file(True)
                .set_show_pending_spinner(False)
            )
        with env_mock_builder.build():
            assert comp_type in [
                CompType.PrefixShown,
                CompType.PrefixHidden,
                CompType.SubsequentHelp,
                CompType.MenuCompletion,
            ]
            env_vars = self.env_vars(
                test_line,
                comp_type,
            )
            client_proc = self.run_client_with_env_vars(
                command_name,
                env_vars,
                expected_exit_code,
            )
            stdout_str = client_proc.stdout.decode("utf-8")
            if expected_stdout_str is not None:
                self.assertEqual(
                    expected_stdout_str,
                    stdout_str,
                )
            stderr_str = client_proc.stderr.decode("utf-8")
            self.assert_no_stderr(
                stderr_str,
                ServerAction.ProposeArgValues,
            )
            return stdout_str

    def assert_DescribeLineArgs(
        self,
        command_name,
        test_line,
        expected_stdout_str: Union[str, None],
        expected_exit_code: Union[int, None] = 0,
        env_mock_builder: EnvMockBuilder = None,
    ):
        if env_mock_builder is None:
            env_mock_builder = (
                EmptyEnvMockBuilder()
                .set_client_config_dict()
                .set_generate_client_config_file(True)
                .set_show_pending_spinner(False)
            )
        with env_mock_builder.build():
            env_vars = self.env_vars(
                test_line,
                CompType.DescribeArgs,
            )
            client_proc = self.run_client_with_env_vars(
                command_name,
                env_vars,
                expected_exit_code,
            )
            stdout_str = client_proc.stdout.decode("utf-8")
            if expected_stdout_str is not None:
                self.assertEqual(
                    expected_stdout_str,
                    stdout_str,
                )
            stderr_str = client_proc.stderr.decode("utf-8")
            self.assert_no_stderr(
                stderr_str,
                ServerAction.ProposeArgValues,
            )
            return ic(stdout_str)

    def assert_RelayLineArgs(
        self,
        command_line_args: list[str],
        expected_stdout_str: Union[str, None],
        expected_stderr_str: Union[str, None],
        expected_exit_code: Union[int, None] = 0,
        env_mock_builder: EnvMockBuilder = None,
    ):
        if env_mock_builder is None:
            env_mock_builder = (
                EmptyEnvMockBuilder()
                .set_client_config_dict()
                .set_generate_client_config_file(True)
                .set_show_pending_spinner(False)
            )
        with env_mock_builder.build():

            client_proc = self.run_client_with_cli_args(
                command_line_args,
                expected_exit_code,
            )
            stdout_str = client_proc.stdout.decode("utf-8")
            if expected_stdout_str is not None:
                self.assertEqual(
                    expected_stdout_str,
                    stdout_str,
                )
            stderr_str = client_proc.stderr.decode("utf-8")
            if expected_stderr_str is not None:
                self.assertEqual(
                    expected_stderr_str,
                    stderr_str,
                )



from enum import (
    auto,
    Enum,
)


class CompositeNodeType(Enum):
    """
    FS_33_76_82_84 composite forest node types.
    """

    zero_arg_node = auto()
    """
    See FS_42_76_93_51 very first zero arg mapping interp.

    It is also related to FS_15_79_76_85 line processor
    (which may potentially be merged with FS_42_76_93_51 zero arg interp).

    TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor
    """

    interp_tree_node = auto()
    """
    See FS_01_89_09_24 interp tree.
    """

    func_tree_node = auto()
    """
    See FS_26_43_73_72 func tree.
    """

    tree_path_node = auto()
    """
    The function of this node is to group other nodes as children under the same tree path.
    """



from __future__ import annotations

from dataclasses import dataclass

from argrelay_lib_root.enum_desc.CompScope import CompScope
from argrelay_lib_root.enum_desc.ServerAction import ServerAction


@dataclass
class UsageStatsEntry:
    """
    Stored entry for FS_87_02_77_34: usage stats.
    """

    server_action: ServerAction

    comp_scope: CompScope

    server_ts_ns: int = 0

    client_version: str = ""

    client_conf_target: str = ""

    client_user_id: str = ""

    command_line: str = ""

    cursor_cpos: int = 0



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)


@dataclass
class ArgValues:
    """
    See also :class:`ArgValuesSchema`.
    """

    arg_values: list[str] = field()
    """
    Values suggested by `ServerAction.ProposeArgValues`.
    """



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)


@dataclass
class SearchControl:
    """
    Implements:
    *   FS_46_96_59_05 `init_control`
    *   FS_31_70_49_15 `search control`

    See also `SearchControlSchema`.
    """

    collection_name: str = field(default_factory = lambda: None)
    """
    MongoDB collection to search `data_envelope`-s in.
    """

    props_to_values_dict: dict[str, str] = field(default_factory = lambda: {})

    # FS_10_93_78_10: `arg_name_to_prop_name_map`:
    # A spec to query `data_envelope`-s for this `EnvelopeContainer`:
    # *   which `prop_name`-s to use in query
    # *   which `arg_name` is mapped into which `prop_name`
    # a `list` to preserve order:
    arg_name_to_prop_name_map: list[dict[str, str]] = field(default_factory = lambda: [])

    # `arg_name_to_prop_name_dict` is derived from `arg_name_to_prop_name_map`:
    # a `dict` for quick lookup:
    arg_name_to_prop_name_dict: dict[str, str] = field(init = False, default_factory = lambda: {})

    # `prop_name_to_arg_name_dict` is derived from `arg_name_to_prop_name_map`:
    # a `dict` for reverse lookup:
    prop_name_to_arg_name_dict: dict[str, str] = field(init = False, default_factory = lambda: {})

    def __post_init__(self):
        self._init_derived_fields()
        self._verify_prop_names_consistency()

    def _init_derived_fields(self):
        # generate `dict`:
        self.arg_name_to_prop_name_dict = self.convert_list_of_ordered_singular_dicts_to_unordered_dict(
            self.arg_name_to_prop_name_map,
        )

        # generate reverse `dict`:
        self.prop_name_to_arg_name_dict = {
            v: k for k, v in
            self.arg_name_to_prop_name_dict.items()
        }

    def _verify_prop_names_consistency(self):
        """
        Make sure `prop_name`-s specified in `props_to_values_dict` for FS_46_96_59_05 `init_control`
        are also among `prop_name`-s specified for FS_31_70_49_15 `search control`.
        """
        for prop_name in self.props_to_values_dict:
            assert prop_name in self.prop_name_to_arg_name_dict

    @staticmethod
    def convert_list_of_ordered_singular_dicts_to_unordered_dict(
        arg_name_to_prop_name_map: list[dict],
    ) -> dict:
        """
        Convert ordered list[dict] (with dict having single { key: value }) into unordered dict { keys: values }.
        """
        converted_dict = {}
        for arg_name_to_prop_name_entry in arg_name_to_prop_name_map:
            arg_name = next(iter(arg_name_to_prop_name_entry))
            prop_name = arg_name_to_prop_name_entry[arg_name]
            converted_dict[arg_name] = prop_name

        return converted_dict



from marshmallow import (
    fields,
    RAISE,
)

from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_server.runtime_data_server_app.GuiBannerConfig import GuiBannerConfig

meter_html_ = "meter_html"
tagline_html_ = "tagline_html"
header_html_ = "header_html"
footer_html_ = "footer_html"


class GuiBannerConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = GuiBannerConfig

    meter_html = fields.String()
    tagline_html = fields.String()
    header_html = fields.String()
    footer_html = fields.String()


gui_banner_config_desc = TypeDesc(
    dict_schema = GuiBannerConfigSchema(),
    ref_name = GuiBannerConfigSchema.__name__,
    dict_example = {
        meter_html_: "",
        tagline_html_: "",
        header_html_: "",
        footer_html_: "",
    },
    default_file_path = "",
)



from marshmallow import (
    fields,
    RAISE,
    Schema,
    ValidationError,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc

ignored_func_ids_list_ = "ignored_func_ids_list"


def validate_tree_node(interp_selector_sub_tree: dict):
    if isinstance(interp_selector_sub_tree, str):
        return
    elif isinstance(interp_selector_sub_tree, dict):
        for sub_tree_node in interp_selector_sub_tree.values():
            validate_tree_node(sub_tree_node)
    else:
        raise ValidationError(f"neither a str nor a dict: {interp_selector_sub_tree}")


# TODO: TODO_40_10_18_32: add custom base to all schemas:
class InterpTreeInterpFactoryConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    # TODO: FS_42_76_93_51: Decide already whether first zero arg interp deserves its future life.
    #       This `ignored_func_ids_list` field is propagated down from
    #       the first zero arg interp (needed for its logic as root plugin instance in the composite forest)
    #       and it also needs to pass validation of this schema (which it converts its config to):
    # TODO: TODO_19_67_22_89: remove `ignored_func_ids_list` - load as `FuncState.fs_unplugged`:
    # It is an error if `func_id` is published by enabled delegator but missing in `func_selector_tree`.
    # To avoid the error, list such `func_id` in `ignored_func_ids_list`.
    ignored_func_ids_list = fields.List(
        fields.String(),
        required = False,
        load_default = [],
    )


tree_path_interp_factory_config_example = {
    ignored_func_ids_list_: [
        "func_id_some_ignored",
    ],
}
tree_path_interp_factory_config_desc = TypeDesc(
    dict_schema = InterpTreeInterpFactoryConfigSchema(),
    ref_name = InterpTreeInterpFactoryConfigSchema.__name__,
    dict_example = tree_path_interp_factory_config_example,
    default_file_path = "",
)



from enum import (
    auto,
    Enum,
)


class ReservedEnvelopeClass(Enum):
    """
    Names of the `data_envelope` classes reserved by `argrelay`.
    """

    ClassUnknown = auto()

    class_function = auto()

    class_help = auto()
    """
    See FS_71_87_33_52 help_hint.
    """

    class_collection = auto()
    """
    See FS_74_69_61_79 get set data envelope.
    """

    class_no_data = auto()
    """
    Class which is supposed to have no `data_envelope`-s loaded.

    See also `SpecialFunc.func_id_no_data`.
    """

    def __str__(self):
        return self.name



from dataclasses import (
    dataclass,
    field,
)


@dataclass
class MongoClientConfig:
    # Example:
    # "mongodb://test:test@localhost/test?authSource=admin"
    client_connection_string: str = field()



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)

from argrelay_schema_config_server.runtime_data_server_plugin.PluginEntry import PluginEntry


@dataclass
class PluginConfig:
    """
    See also `PluginConfigSchema`.
    """

    server_plugin_instances: dict[str, PluginEntry] = field()
    """
    Key = `plugin_instance_id`
    """

    plugin_instance_id_activate_list: list[str] = field()
    """
    List of `plugin_instance_id`-s in order of activation generated by
    walking through a DAG described by `plugin_dependencies`.
    Each `plugin_instance_id` is a key into `server_plugin_instances`.
    """



from __future__ import annotations

from marshmallow import RAISE

from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema


class BasePluginConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True


def serialize_dag_to_list(
    entire_dag: dict[str, list[str]],
) -> list[str]:
    output_list: list[str] = list_sub_dag(
        [],
        entire_dag,
        [node_id for node_id in entire_dag],
    )
    return output_list


def list_sub_dag(
    curr_path: list[str],
    entire_dag: dict[str, list[str]],
    id_sub_list: list[str],
) -> list[str]:
    output_list: list[str] = []

    for node_id in id_sub_list:
        if node_id in curr_path:
            raise ValueError(f"cyclic ref to plugin id in path `{curr_path}` -> `{node_id}`")
        if node_id not in entire_dag:
            raise ValueError(f"plugin id in path `{curr_path}` -> `{node_id}` is not defined")
        if node_id in entire_dag:
            # plugin has dependencies:
            sub_list = list_sub_dag(
                curr_path + [node_id],
                entire_dag,
                entire_dag[node_id],
            )
            for sub_node_id in sub_list:
                if sub_node_id not in output_list:
                    output_list.append(sub_node_id)
        else:
            raise ValueError(f"plugin id in path `{curr_path}` -> `{node_id}` is not included for activation")

        if node_id not in output_list:
            output_list.append(node_id)

    return output_list



import os

import re

# The "distribution root" refers to the top-level directory where your project code resides,
# is the root directory that contains the `setup.py` file itself.
# In case of `argrelay`, it may confusingly appear it is equivalent to `argrelay_dir`
# (because it contains `setup.py`), but it is not - when installed, `setup.py` will run from extracted archive:
distrib_root = os.path.dirname(os.path.abspath(__file__))

# Implements this:
# https://stackoverflow.com/a/7071358/441652
version_file = f"{distrib_root}/src/argrelay/_version.py"
version_content = open(version_file, "rt").read()
version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
regex_match = re.search(version_regex, version_content, re.M)
if regex_match:
    version_string = regex_match.group(1)
else:
    raise RuntimeError(f"Unable to find version string: ${version_file}")

import setuptools

tests_require = [
    "tox",
    "responses",
    "mongomock",
    "pandas",
    "icecream",
    "jsonpath-ng",
    "import-linter",
]

# To install these extra dev dependencies:
# pip install --editable "${argrelay_dir}/"[tests]
extras_require = {
    "tests": tests_require,
}


def list_dir(
    top_dir_abs_path,
) -> list[str]:
    """
    List files recursively from `top_dir_abs_path` with paths relative to `top_dir_abs_path`.
    """
    file_rel_paths = []
    for (parent_dir_abs_path, child_dir_names, child_file_names) in os.walk(top_dir_abs_path):
        for child_file_name in child_file_names:
            file_abs_path = os.path.join(
                parent_dir_abs_path,
                child_file_name,
            )
            file_rel_path = os.path.relpath(
                file_abs_path,
                top_dir_abs_path,
            )
            file_rel_paths.append(file_rel_path)
    return file_rel_paths


def prefix_file_rel_paths(
    prefix_rel_path: str,
    file_rel_paths: list[str],
) -> list[str]:
    """
    Prefix every given path in `file_rel_paths` with `prefix_rel_path`.
    """
    file_prefixed_rel_paths = []
    for file_rel_path in file_rel_paths:
        file_prefixed_rel_path = os.path.join(
            prefix_rel_path,
            file_rel_path,
        )
        file_prefixed_rel_paths.append(file_prefixed_rel_path)
    return file_prefixed_rel_paths


argrelay_docs_files = prefix_file_rel_paths(
    "./docs/",
    list_dir(f"{distrib_root}/docs/"),
) + prefix_file_rel_paths(
    "./",
    [
        "readme.md",
    ],
)

argrelay_data_files = prefix_file_rel_paths(
    "./data/",
    list_dir(f"{distrib_root}/data/"),
)

setuptools.setup(
    name = "argrelay",
    version = version_string,
    author = "uvsmtid",
    author_email = "uvsmtid@gmail.com",
    description = "A data server to CLI tools with attribute search & Tab-completion in Bash shell",
    long_description = """
See: https://github.com/argrelay/argrelay
    """,
    long_description_content_type = "text/markdown",
    keywords = "argparse, argcomplete, bash, complete",
    url = "https://github.com/argrelay/argrelay",
    project_urls = {
        "Bug Tracker": "https://github.com/argrelay/argrelay/issues",
    },
    classifiers = [
        "Development Status :: 3 - Alpha",

        "Environment :: Console",
        "Framework :: Flask",
        "Intended Audience :: Information Technology",

        "Topic :: Terminals",
        "Topic :: System :: Shells",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Text Processing :: Indexing",

        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",

        "Programming Language :: Python :: 3",

        "License :: OSI Approved :: Apache Software License",

        "Operating System :: POSIX :: Linux",
    ],
    # See sample layout:
    # https://docs.python.org/3.8/distutils/setupscript.html#installing-package-data
    # List all packages/sub-packages (so that they are taken by `package_dir` below):
    packages = setuptools.find_packages(
        where = f"{distrib_root}/src/",
    ) + [
        "argrelay_docs",
        "argrelay_data",
    ],
    # See:
    # https://docs.python.org/3.8/distutils/setupscript.html#listing-whole-packages
    #     The keys to this dictionary are package names,
    #     and an empty package name stands for the root package.
    #     The values are directory names relative to your distribution root.
    #     See "distribution root" above - during
    package_dir = {
        "argrelay":
            "./src/argrelay/",
        "argrelay_api_plugin_abstract":
            "./src/argrelay_api_plugin_abstract/",
        "argrelay_api_plugin_check_env_abstract":
            "./src/argrelay_api_plugin_check_env_abstract/",
        "argrelay_api_plugin_client_abstract":
            "./src/argrelay_api_plugin_client_abstract/",
        "argrelay_api_plugin_server_abstract":
            "./src/argrelay_api_plugin_server_abstract/",
        "argrelay_api_server_cli":
            "./src/argrelay_api_server_cli/",
        "argrelay_app_check_env":
            "./src/argrelay_app_check_env/",
        "argrelay_app_bootstrap":
            "./src/argrelay_app_bootstrap/",
        "argrelay_app_client":
            "./src/argrelay_app_client/",
        "argrelay_app_server":
            "./src/argrelay_app_server/",
        "argrelay_docs":
            "./",
        "argrelay_data":
            "./",
        "argrelay_lib_check_env_plugin_core":
            "./src/argrelay_lib_check_env_plugin_core/",
        "argrelay_lib_root":
            "./src/argrelay_lib_root/",
        "argrelay_lib_server_plugin_check_env":
            "./src/argrelay_lib_server_plugin_check_env/",
        "argrelay_lib_server_plugin_core":
            "./src/argrelay_lib_server_plugin_core/",
        "argrelay_lib_server_plugin_demo":
            "./src/argrelay_lib_server_plugin_demo/",
        "argrelay_schema_config_check_env":
            "./src/argrelay_schema_config_check_env/",
        "argrelay_schema_config_client":
            "./src/argrelay_schema_config_client/",
        "argrelay_schema_config_server":
            "./src/argrelay_schema_config_server/",
        "argrelay_test_infra":
            "./src/argrelay_test_infra/",
    },
    # See:
    # https://docs.python.org/3.8/distutils/setupscript.html#installing-package-data
    #     The paths are interpreted as relative to the directory containing the package
    #     (information from the `package_dir` mapping is used if appropriate);
    #     that is, the files are expected to be part of the package in the source directories.
    package_data = {
        "argrelay": [
            # Other resource files:
            "custom_integ_res/argrelay_common_lib.bash",
            "custom_integ_res/shell_env.bash",
            "custom_integ_res/bootstrap_env.bash",
            "custom_integ_res/check_env.bash",
            "custom_integ_res/dev_shell.bash",
            "custom_integ_res/init_shell_env.bash",
            "custom_integ_res/upgrade_env_packages.bash",

            # Files in `script_plugin.d`:
            "custom_integ_res/script_plugin.d/check_env_plugin.all_argrelay_plugins.bash",
            "custom_integ_res/script_plugin.d/check_env_plugin.bash_version.bash",
            "custom_integ_res/script_plugin.d/check_env_plugin.git_version.bash",

        ],
        "argrelay_app_bootstrap": [
            # Config files:
            "sample_conf/argrelay_client.json",
            "sample_conf/argrelay_server.yaml",
            "sample_conf/argrelay_plugin.yaml",
            "sample_conf/check_env_plugin.conf.bash",
            "sample_conf/check_env_plugin.conf.yaml",
        ],
        "argrelay_app_server": [
            # GUI client:
            "argrelay_app_server/relay_server/gui_static/argrelay_client.js",
            "argrelay_app_server/relay_server/gui_static/argrelay_favicon_16.ico",
            "argrelay_app_server/relay_server/gui_static/argrelay_style.css",
            "argrelay_app_server/relay_server/gui_static/external_link.svg",
            "argrelay_app_server/relay_server/gui_templates/argrelay_main.html",
        ],
        "argrelay_docs": argrelay_docs_files,
        "argrelay_data": argrelay_data_files,
    },
    include_package_data = True,
    # Because `argrelay_docs` and `argrelay_data` share the same `package_dir` config,
    # exclude overlapping files explicitly (opposite of `package_data` config):
    exclude_package_data = {
        "argrelay_docs": argrelay_data_files,
        "argrelay_data": argrelay_docs_files,
    },
    # FS_84_11_73_28: supported python versions:
    python_requires = ">=3.9",
    install_requires = [
        "Flask",
        "Werkzeug",
        "PyYaml",
        "jsonschema",
        "flasgger",
        "marshmallow<4.0.0",
        "marshmallow-oneofschema",
        "apispec",
        "pymongo",
        "GitPython",
        # Use `mongomock` as replacement for Mongo DB in simple prod cases:
        "mongomock",
        "requests",
        "cachetools",
    ],
    extras_require = extras_require,
)



from __future__ import annotations

import json

from argrelay_api_plugin_server_abstract.delegator_utils import redirect_to_not_disambiguated_error
from argrelay_api_plugin_server_abstract.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_root.enum_desc.SpecialFunc import SpecialFunc
from argrelay_lib_server_plugin_core.plugin_delegator.client_invocation_utils import prohibit_unconsumed_args
from argrelay_lib_server_plugin_core.plugin_delegator.DelegatorDataBackendBase import (
    DelegatorDataBackendBase,
    get_collection_search_control,
)
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    instance_data_
)
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)

collection_name_container_ipos_ = 1
data_envelope_container_ipos_ = 2


class DelegatorDataBackendGet(DelegatorDataBackendBase):
    """
    Implements `SpecialFunc.func_id_get_data_envelopes` for FS_74_69_61_79 get set data envelope.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        collection_search_control = get_collection_search_control(
        )

        func_envelopes = []

        given_function_envelope = {
            instance_data_: {
                func_id_: SpecialFunc.func_id_get_data_envelopes.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    collection_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
            ReservedPropName.help_hint.name: "Get `data_envelope`-s based on their `index_prop`-s.",
            ReservedPropName.func_state.name: FuncState.fs_alpha.name,
            ReservedPropName.func_id.name: SpecialFunc.func_id_get_data_envelopes.name,
        }
        func_envelopes.append(given_function_envelope)

        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:

        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == SpecialFunc.func_id_get_data_envelopes.name

        vararg_container_ipos = data_envelope_container_ipos_

        # Verify that func is selected:
        if interp_ctx.curr_container_ipos < vararg_container_ipos:
            return redirect_to_not_disambiguated_error(
                interp_ctx,
                local_server.plugin_config,
                ReservedEnvelopeClass.class_collection.name,
            )

        vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]

        # All what is left to do is to query objects.
        # Search `data_envelope`-s based on existing args on command line:
        vararg_container.data_envelopes = (
            local_server
            .get_query_engine()
            .query_data_envelopes_for(vararg_container)
        )

        # Plugin to invoke on client side:
        delegator_plugin_instance_id = self.plugin_instance_id
        # Package into `InvocationInput` payload object:
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                delegator_plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        func_id = get_func_id_from_invocation_input(invocation_input)
        assert func_id == SpecialFunc.func_id_get_data_envelopes.name

        prohibit_unconsumed_args(invocation_input)

        for data_envelope in invocation_input.envelope_containers[data_envelope_container_ipos_].data_envelopes:
            print(json.dumps(data_envelope))



from argrelay_app_client.client_pipeline.BytesHandlerAbstract import BytesHandlerAbstract
from argrelay_app_client.client_pipeline.BytesSrcAbstract import BytesSrcAbstract


class BytesSrcLocal(BytesSrcAbstract):

    def __init__(
        self,
        bytes_handler: BytesHandlerAbstract,
    ):
        super().__init__()
        self.bytes_handler = bytes_handler

    def accept_bytes(
        self,
        given_bytes: bytes,
    ):
        self.bytes_handler.handle_bytes(given_bytes)



from __future__ import annotations

from marshmallow import (
    fields,
    RAISE,
    Schema,
)

from argrelay_api_server_cli.schema_response.EnvelopeContainerSchema import data_envelopes_
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    data_envelope_desc,
    sample_prop_name_a_,
    sample_prop_name_b_,
    sample_prop_name_c_,
)

envelope_class_to_collection_name_map_ = "envelope_class_to_collection_name_map"
collection_name_to_index_props_map_ = "collection_name_to_index_props_map"


class ConfigOnlyLoaderConfigSchema(Schema):
    """
    Part of FS_49_96_50_77 config_only_loader implementation.
    """

    class Meta:
        unknown = RAISE
        strict = True

    # Specifies `collection_name` for each `envelope_class` to load.
    # By default, if mapping for `envelope_class` is not specified,
    # the same `collection_name` is used as `envelope_class`.
    envelope_class_to_collection_name_map = fields.Dict(
        keys = fields.String(),
        values = fields.String(),
        required = False,
        load_default = {},
    )

    collection_name_to_index_props_map = fields.Dict(
        keys = fields.String(),
        values = fields.List(
            fields.String(),
            required = True,
        ),
        required = True,
    )

    data_envelopes = fields.List(
        fields.Nested(data_envelope_desc.dict_schema),
        required = True,
    )


config_only_loader_config_desc = TypeDesc(
    dict_schema = ConfigOnlyLoaderConfigSchema(),
    ref_name = ConfigOnlyLoaderConfigSchema.__name__,
    dict_example = {
        envelope_class_to_collection_name_map_: {
            ReservedEnvelopeClass.class_function.name: ReservedEnvelopeClass.class_function.name,
        },
        collection_name_to_index_props_map_: {
            ReservedEnvelopeClass.class_function.name: [
                sample_prop_name_a_,
                sample_prop_name_b_,
                sample_prop_name_c_,
            ],
        },
        data_envelopes_: [
            data_envelope_desc.dict_example,
        ],
    },
    default_file_path = "",
)



from argrelay_api_plugin_abstract.AbstractPlugin import (
    AbstractPlugin,
    import_plugin_class,
)


class PluginClientAbstract(AbstractPlugin):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict,
        )


def instantiate_client_plugin(
    plugin_instance_id: str,
    plugin_entry,
):
    plugin_class = import_plugin_class(plugin_entry)
    plugin_object: AbstractPlugin = plugin_class(
        plugin_instance_id,
        plugin_entry.plugin_config,
    )
    return plugin_object



from enum import (
    auto,
    Enum,
)


class ServiceEnvelopeClass(Enum):
    class_cluster = auto()
    class_host = auto()
    class_service = auto()
    class_access_type = auto()

    def __str__(self):
        return self.name



from argrelay_api_plugin_check_env_abstract.CheckEnvResult import CheckEnvResult
from argrelay_lib_check_env_plugin_core.PluginCheckEnvServerResponseValueAbstract import (
    PluginCheckEnvServerResponseValueAbstract,
)
from argrelay_lib_check_env_plugin_core.SchemaPluginCheckEvnServerResponseValueAbstract import (
    field_values_to_command_lines_,
)
from argrelay_lib_root.enum_desc.CheckEnvField import CheckEnvField
from argrelay_lib_root.enum_desc.ResultCategory import ResultCategory
from argrelay_lib_root.misc_helper_common import get_argrelay_dir
from argrelay_lib_server_plugin_demo.demo_git.git_utils import (
    get_full_commit_id,
    is_git_repo,
)

_commit_id_display_length = 8


class PluginCheckEnvServerResponseValueCommitId(PluginCheckEnvServerResponseValueAbstract):

    def __init__(
        self,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            plugin_instance_id,
            plugin_config_dict = plugin_config_dict or {
                field_values_to_command_lines_: {
                    CheckEnvField.server_git_commit_id.name: "argrelay.check_env server_commit",
                },
            },
        )

    def verify_online_value(
        self,
        field_name: str,
        field_value,
    ) -> CheckEnvResult:
        argrelay_dir = get_argrelay_dir()

        if is_git_repo(argrelay_dir):
            try:
                client_commit_id = get_full_commit_id(argrelay_dir)
            except ValueError:
                client_commit_id = None
        else:
            client_commit_id = None

        if client_commit_id is None:
            return CheckEnvResult(
                result_category = ResultCategory.ExecutionFailure,
                result_key = field_name,
                result_value = field_value,
                result_message = "Failed on getting client commit id",
            )
        else:
            if client_commit_id == field_value:
                return CheckEnvResult(
                    result_category = ResultCategory.VerificationSuccess,
                    result_key = field_name,
                    result_value = field_value[:_commit_id_display_length],
                    result_message = f"The runtime server commit id matches client one [{client_commit_id[:_commit_id_display_length]}]",
                )
            else:
                return CheckEnvResult(
                    result_category = ResultCategory.VerificationWarning,
                    result_key = field_name,
                    result_value = field_value[:_commit_id_display_length],
                    result_message = f"The runtime server commit id does not match client one [{client_commit_id[:_commit_id_display_length]}]",
                )



from __future__ import annotations

from argrelay_api_plugin_server_abstract.delegator_utils import get_config_value_once
from argrelay_api_plugin_server_abstract.DelegatorAbstract import (
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.CheckEnvField import CheckEnvField
from argrelay_lib_root.enum_desc.CheckEnvFunc import CheckEnvFunc
from argrelay_lib_root.enum_desc.FuncState import FuncState
from argrelay_lib_root.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_check_env.DelegatorCheckEnvBase import DelegatorCheckEnvBase
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import instance_data_
from argrelay_schema_config_server.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    func_id_,
    search_control_list_,
)


class DelegatorCheckEnvGitCommitId(DelegatorCheckEnvBase):
    """
    Implements `CheckEnvFunc.func_id_get_server_project_git_commit_id` for FS_36_17_84_44 `check_env`.
    """

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:
        func_envelopes = [
            {
                instance_data_: {
                    func_id_: CheckEnvFunc.func_id_get_server_project_git_commit_id.name,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.class_function.name,
                ReservedPropName.help_hint.name: (
                    CheckEnvFunc.func_id_get_server_project_git_commit_id.name
                ),
                ReservedPropName.func_state.name: FuncState.fs_alpha.name,
                ReservedPropName.func_id.name: CheckEnvFunc.func_id_get_server_project_git_commit_id.name,
            },
        ]
        return func_envelopes

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        func_id = get_func_id_from_interp_ctx(interp_ctx)
        assert func_id == CheckEnvFunc.func_id_get_server_project_git_commit_id.name

        project_git_commit_id: str = get_config_value_once(
            self.server_config.server_configurators,
            lambda server_configurator: server_configurator.provide_project_git_commit_id(),
            None,
        )
        custom_plugin_data = {
            CheckEnvField.server_git_commit_id.name: project_git_commit_id,
        }

        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                self.plugin_instance_id
            ],
            custom_plugin_data = custom_plugin_data,
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        func_id = get_func_id_from_invocation_input(invocation_input)
        assert func_id == CheckEnvFunc.func_id_get_server_project_git_commit_id.name

        print(f"{invocation_input.custom_plugin_data[CheckEnvField.server_git_commit_id.name]}")



from __future__ import annotations

from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext


class DelegatorNoopBase(DelegatorSingleFuncAbstract):
    """
    Base delegator for all delegators doing nothing.
    """

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                self.plugin_instance_id
            ],
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        # Do nothing:
        pass



from __future__ import annotations

from argrelay_api_plugin_server_abstract.delegator_utils import (
    redirect_to_not_disambiguated_error,
    set_default_to,
)
from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.ReservedPropName import ReservedPropName
from argrelay_lib_server_plugin_demo.demo_service.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay_lib_server_plugin_demo.demo_service.ServicePropName import ServicePropName
from argrelay_schema_config_server.schema_config_interp.SearchControlSchema import populate_search_control


def get_access_search_control(
) -> dict:
    return populate_search_control(
        ServiceEnvelopeClass.class_access_type.name,
        {
            ReservedPropName.envelope_class.name: ServiceEnvelopeClass.class_access_type.name,
        },
        [
            # TODO: TODO_61_99_68_90: figure out what to do with explicit `envelope_class` `search_prop`:
            {"class": ReservedPropName.envelope_class.name},

            {"access": ServicePropName.access_type.name},
        ],
    )


class DelegatorServiceBase(DelegatorSingleFuncAbstract):

    def _fill_object_container(
        self,
        any_assignment,
        interp_ctx,
        object_container_ipos,
    ):
        if (
            # TODO: TODO_73_23_85_93: use helper to select container ipos:
            interp_ctx.curr_container_ipos == interp_ctx.curr_interp.base_container_ipos + object_container_ipos
        ):
            service_container = interp_ctx.envelope_containers[(
                # TODO: TODO_73_23_85_93: use helper to select container ipos:
                interp_ctx.curr_interp.base_container_ipos + object_container_ipos
            )]
            any_assignment = (
                set_default_to(ServicePropName.run_mode.name, "active", service_container)
                or
                any_assignment
            )
        return any_assignment

    def _fill_access_container(
        self,
        any_assignment,
        interp_ctx,
        object_container_ipos,
        access_container_ipos,
    ):
        # If we need to specify `access_type` `data_envelope`:
        # TODO: TODO_73_23_85_93: use helper to select container ipos:
        if interp_ctx.curr_container_ipos == interp_ctx.curr_interp.base_container_ipos + access_container_ipos:
            # Take object found so far:
            data_envelope = interp_ctx.envelope_containers[(
                # TODO: TODO_73_23_85_93: use helper to select container ipos:
                interp_ctx.curr_interp.base_container_ipos + object_container_ipos
            )].data_envelopes[0]

            access_container = interp_ctx.envelope_containers[(
                # TODO: TODO_73_23_85_93: use helper to select container ipos:
                interp_ctx.curr_interp.base_container_ipos + access_container_ipos
            )]

            # Select default value to search `access_type` `data_envelope` based on `code_maturity`:
            code_prop_name = ServicePropName.code_maturity.name
            if code_prop_name in data_envelope:
                code_arg_val = data_envelope[code_prop_name]
                if code_arg_val == "prod":
                    any_assignment = (
                        set_default_to(ServicePropName.access_type.name, "ro", access_container)
                        or
                        any_assignment
                    )
                else:
                    any_assignment = (
                        set_default_to(ServicePropName.access_type.name, "rw", access_container)
                        or
                        any_assignment
                    )

        return any_assignment

    def _compose_invocation_input_for_list(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
        vararg_container_ipos,
        envelope_class,
    ):
        # Verify that func is selected and all what is left to do is to query 0...N objects:
        if interp_ctx.curr_container_ipos >= vararg_container_ipos:
            # Search `data_envelope`-s based on existing args on command line:
            vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]
            vararg_container.data_envelopes = (
                local_server
                .get_query_engine()
                .query_data_envelopes_for(vararg_container)
            )

            # Plugin to invoke on client side:
            delegator_plugin_instance_id = self.plugin_instance_id
            # Package into `InvocationInput` payload object:
            invocation_input = InvocationInput.with_interp_context(
                interp_ctx,
                delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
                    delegator_plugin_instance_id
                ],
                custom_plugin_data = {},
            )
            return invocation_input
        else:
            return redirect_to_not_disambiguated_error(
                interp_ctx,
                local_server.plugin_config,
                envelope_class,
            )



from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_client.relay_client import ClientCommandAbstract
from argrelay_app_client.relay_client.ClientCommandFactoryAbstract import ClientCommandFactoryAbstract
from argrelay_lib_root.misc_helper_common.ElapsedTime import ElapsedTime
from argrelay_schema_config_client.runtime_data_client_app.ClientConfig import ClientConfig


class ClientAbstract:
    """
    An abstract `relay_client` to `relay_server`
    """

    def __init__(
        self,
        client_config: ClientConfig,
        command_factory: ClientCommandFactoryAbstract
    ):
        self.client_config: ClientConfig = client_config
        self.command_factory: ClientCommandFactoryAbstract = command_factory

    # noinspection PyMethodMayBeStatic
    def make_request(
        self,
        call_ctx: CallContext,
    ) -> ClientCommandAbstract:
        command_obj = self.command_factory.create_command(call_ctx)
        command_obj.execute_command()
        ElapsedTime.measure("after_execute_command")
        return command_obj



from argrelay_api_plugin_server_abstract.DelegatorSingleFuncAbstract import DelegatorSingleFuncAbstract
from argrelay_api_server_cli.schema_response.InvocationInput import InvocationInput
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.ClientExitCode import ClientExitCode
from argrelay_lib_root.misc_helper_common import eprint
from argrelay_lib_server_plugin_core.plugin_delegator.SchemaCustomDataDelegatorError import (
    error_code_,
    error_message_,
)


class DelegatorError(DelegatorSingleFuncAbstract):
    """
    This delegator is routed to and invoked on the client side to report an error.

    There is no error `func_id` for DelegatorError (unlike for many other delegators), at least for now.
    Instead, other delegators can override themselves and provide error info to use it by the client.
    """

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        delegator_plugin_entry = local_server.plugin_config.server_plugin_instances[
            self.plugin_instance_id
        ]
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = delegator_plugin_entry,
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ) -> None:
        error_message = "ERROR: unknown error"
        error_code = ClientExitCode.GeneralError.value
        if invocation_input.custom_plugin_data:
            if error_message_ in invocation_input.custom_plugin_data:
                error_message = invocation_input.custom_plugin_data[error_message_]
            if error_code_ in invocation_input.custom_plugin_data:
                error_code = invocation_input.custom_plugin_data[error_code_]
        eprint(error_message)
        exit(error_code)



from argrelay_api_server_cli.schema_response.ArgValuesSchema import arg_values_
from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_server.handler_request.AbstractServerRequestHandler import AbstractServerRequestHandler
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_lib_root.enum_desc.ServerAction import ServerAction
from argrelay_lib_root.misc_helper_common.ElapsedTime import ElapsedTime


class ProposeArgValuesServerRequestHandler(AbstractServerRequestHandler):

    def __init__(
        self,
        local_server: LocalServer,
    ):
        super().__init__(
            local_server = local_server,
        )

    def handle_request(
        self,
        call_ctx: CallContext,
    ) -> dict:
        assert call_ctx.server_action is ServerAction.ProposeArgValues
        self._store_usage_stats_entry(call_ctx)

        self.interpret_command(self.local_server, call_ctx)
        ElapsedTime.measure("after_interpret_command")

        return {
            arg_values_: self.interp_ctx.propose_arg_values()
        }



from __future__ import annotations

from copy import deepcopy

import mongomock
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database

from argrelay_app_server.mongo_data.ProgressTracker import ProgressTracker
from argrelay_schema_config_server.runtime_data_server_app.EnvelopeCollection import EnvelopeCollection
from argrelay_schema_config_server.runtime_data_server_app.MongoConfig import MongoConfig
from argrelay_schema_config_server.schema_config_interp.DataEnvelopeSchema import (
    data_envelope_desc,
    envelope_id_,
    mongo_id_,
)


def get_mongo_client(mongo_config: MongoConfig):
    if mongo_config.use_mongomock:
        return mongomock.MongoClient()
    else:
        return MongoClient(mongo_config.mongo_client.client_connection_string)


def store_envelopes(
    mongo_db: Database,
    cleaned_mongo_collections: set[str],
    envelope_collections: list[EnvelopeCollection],
    progress_tracker: ProgressTracker,
):
    # Calculate total:
    for envelope_collection in envelope_collections:
        progress_tracker.total_envelope_n += len(envelope_collection.data_envelopes)

    # Index all:
    for envelope_collection in envelope_collections:
        store_envelope_collection(
            mongo_db,
            cleaned_mongo_collections,
            envelope_collection,
            progress_tracker,
        )

    assert progress_tracker.total_envelope_i == progress_tracker.total_envelope_n


def delete_data_envelopes(
    mongo_db: Database,
    collection_name: str,
    query_dict: dict,
) -> None:
    col_proxy: Collection = mongo_db[collection_name]
    col_proxy.delete_many(query_dict)


def store_envelope_collection(
    mongo_db: Database,
    cleaned_mongo_collections: set[str],
    envelope_collection: EnvelopeCollection,
    progress_tracker: ProgressTracker,
) -> None:
    collection_name = envelope_collection.collection_name
    col_proxy: Collection = mongo_db[collection_name]
    if collection_name not in cleaned_mongo_collections:
        col_proxy.delete_many({})
        col_proxy.drop_indexes()
        cleaned_mongo_collections.add(collection_name)

    progress_tracker.track_collection_indexing_start(
        collection_name,
        len(envelope_collection.data_envelopes),
    )

    for data_envelope in envelope_collection.data_envelopes:

        progress_tracker.track_collection_indexing_increment(collection_name)
        envelope_to_store = deepcopy(data_envelope)

        try:
            data_envelope_desc.validate_dict(envelope_to_store)

            if envelope_id_ in envelope_to_store:
                envelope_to_store[mongo_id_] = data_envelope[envelope_id_]

            col_proxy.insert_one(envelope_to_store)
        except:
            print(f"envelope_to_store: {envelope_to_store}")
            # Rethrow previous error:
            raise

    progress_tracker.track_collection_indexing_stop(collection_name)


def create_index(
    mongo_db: Database,
    collection_name: str,
    index_props: list[str],
):
    col_proxy: Collection = mongo_db[collection_name]

    for index_prop in index_props:
        col_proxy.create_index(index_prop)



import time

from argrelay_api_server_cli.server_spec.CallContext import CallContext
from argrelay_app_server.relay_server.LocalServer import LocalServer
from argrelay_app_server.relay_server.UsageStatsEntry import UsageStatsEntry
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_app_server.runtime_context.ParsedContext import ParsedContext


class AbstractServerRequestHandler:

    def __init__(
        self,
        local_server: LocalServer,
    ):
        self.local_server: LocalServer = local_server
        self.interp_ctx: InterpContext

    def handle_request(
        self,
        call_ctx: CallContext,
    ) -> dict:
        raise NotImplementedError

    def interpret_command(
        self,
        local_server: LocalServer,
        call_ctx: CallContext,
    ) -> None:
        parsed_ctx = ParsedContext.from_instance(call_ctx)
        # TODO: Split server_config and static_data in argrelay_server.yaml (both top level, not config including data):
        self.interp_ctx = InterpContext(
            parsed_ctx = parsed_ctx,
            interp_factories = local_server.server_config.interp_factories,
            action_delegators = local_server.server_config.action_delegators,
            query_engine = local_server.get_query_engine(),
            help_hint_cache = local_server.help_hint_cache,
        )
        self.interp_ctx.interpret_command(local_server.server_config.server_plugin_control.first_interp_factory_id)
        self.interp_ctx.print_debug()

    @staticmethod
    def _create_usage_stats_entry(
        call_ctx: CallContext,
    ) -> UsageStatsEntry:
        usage_stats_entry = UsageStatsEntry(
            server_action = call_ctx.server_action,
            comp_scope = call_ctx.comp_scope,
            server_ts_ns = time.time_ns(),
            client_version = call_ctx.client_version,
            client_conf_target = call_ctx.client_conf_target,
            client_user_id = call_ctx.client_uid,
            command_line = call_ctx.command_line,
            cursor_cpos = call_ctx.cursor_cpos,
        )
        return usage_stats_entry

    def _store_usage_stats_entry(
        self,
        call_ctx: CallContext,
    ) -> None:
        usage_stats_entry = self._create_usage_stats_entry(
            call_ctx = call_ctx,
        )
        self.local_server.usage_stats_store.store_usage_stats_entry(
            usage_stats_entry,
        )



class BytesSrcAbstract:

    def __init__(
        self,
    ):
        super().__init__()

    def accept_bytes(
        self,
        given_bytes: bytes,
    ):
        # Implement in derived class:
        raise NotImplementedError()



from marshmallow import (
    fields,
    RAISE,
)

from argrelay_lib_root.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc
from argrelay_schema_config_server.runtime_data_server_app.MongoClientConfig import MongoClientConfig

client_connection_string_ = "client_connection_string"


class MongoClientConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        strict = True

    model_class = MongoClientConfig

    client_connection_string = fields.String()
    """
    See: https://www.mongodb.com/docs/manual/reference/connection-string/

    Full format:
    mongodb://[username:password@]host1[:port1][,...hostN[:portN]][/[defaultauthdb][?options]]

    The simplest:
    mongodb://localhost
    """


mongo_client_config_desc = TypeDesc(
    dict_schema = MongoClientConfigSchema(),
    ref_name = MongoClientConfigSchema.__name__,
    dict_example = {
        client_connection_string_: "mongodb://localhost",
    },
    default_file_path = "",
)



from enum import (
    auto,
    Enum,
)


class ReservedPropName(Enum):
    """
    Some of the `data_envelope` props used to implement various `argrelay` features.
    """

    collection_name = auto()
    """
    Normally, `collection_name` is not part of `data_envelope` - instead, it is only used in MongoDB query API.
    But it is part of `data_envelope` (see FS_74_69_61_79 get set data envelope)
    with `ReservedEnvelopeClass.class_collection` `envelope_class`.
    """

    # TODO: Maybe rename to `envelope_class` everywhere for consistency?
    envelope_class = auto()
    """
    See also synonym `class_name`.
    """

    func_id = auto()
    """
    Used by FS_26_43_73_72 func tree to distinguish leaves (func ids).
    It is normally the same with `FunctionEnvelopeInstanceDataSchema.func_id_`.
    """

    prop_name = auto()

    prop_value = auto()

    help_hint = auto()

    func_state = auto()



import subprocess
from subprocess import (
    Popen,
    TimeoutExpired,
)

from argrelay_lib_root.misc_helper_common import eprint
from argrelay_schema_config_server.runtime_data_server_app.MongoConfig import MongoConfig


class MongoServerWrapper:

    def __init__(self):
        self.use_mongomock: bool = True
        self.is_started: bool = False
        self.mongo_proc: Popen

    def start_mongo_server(self, mongo_config: MongoConfig):

        eprint(f"use_mongomock: {mongo_config.use_mongomock}")
        if mongo_config.use_mongomock:
            self.use_mongomock = True
            return
        else:
            self.use_mongomock = False

        eprint(f"start_server: {mongo_config.mongo_server.start_server}")
        eprint(f"is_started: {self.is_started}")
        if mongo_config.mongo_server.start_server and not self.is_started:
            eprint(f"server_start_command: {mongo_config.mongo_server.server_start_command}")
            self.mongo_proc = subprocess.Popen(mongo_config.mongo_server.server_start_command, shell = True)
            try:
                self.mongo_proc.wait(timeout = 5)
            except TimeoutExpired:
                # Ignore: it probably started without issues. Or did it?
                pass
            exit_code = self.mongo_proc.poll()
            eprint("mongo_proc: exit_code: ", exit_code)
            if not exit_code:
                # It must be still running:
                self.is_started = True
            elif exit_code != 0:
                # Something went wrong:
                raise RuntimeError

    def stop_mongo_server(self):

        if self.use_mongomock:
            return

        if self.is_started:
            self.mongo_proc.kill()
        while self.is_started:
            try:
                self.mongo_proc.wait(timeout = 5)
            except TimeoutExpired:
                eprint("still running: mongo_proc: ", self.mongo_proc)
            exit_code = self.mongo_proc.poll()
            eprint("mongo_proc: exit_code: ", exit_code)
            if exit_code:
                # Process finished:
                self.is_started = False



from __future__ import annotations

import contextlib
from copy import deepcopy
from typing import Union

from argrelay_app_server.composite_forest.CompositeForestExtractor import extract_interp_tree
from argrelay_app_server.composite_forest.CompositeInfoType import CompositeInfoType
from argrelay_app_server.composite_forest.DictTreeWalker import (
    DictTreeWalker,
    fetch_subtree_node,
    sequence_starts_with,
    surrogate_node_id_,
)
from argrelay_app_server.misc_helper_server import insert_unique_to_sorted_list
from argrelay_app_server.runtime_context.AbstractArg import ArgCommandValueOffered
from argrelay_app_server.runtime_context.InterpContext import InterpContext
from argrelay_lib_root.enum_desc.InterpStep import InterpStep
from argrelay_lib_root.enum_desc.PluginType import PluginType
from argrelay_lib_server_plugin_core.plugin_interp.AbstractInterpFactory import (
    AbstractInterp,
    AbstractInterpFactory,
)
from argrelay_lib_server_plugin_core.plugin_interp.InterpTreeInterpFactoryConfigSchema import (
    tree_path_interp_factory_config_desc,
)
from argrelay_lib_server_plugin_core.plugin_interp.NoopInterpFactory import NoopInterpFactory
from argrelay_schema_config_server.runtime_data_server_app.ServerConfig import (
    assert_plugin_instance_id,
    ServerConfig,
)


class InterpTreeInterpFactory(AbstractInterpFactory):
    """
    Implements FS_01_89_09_24 interp tree.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )
        self.is_recursive_func_load = False
        self.interp_selector_tree = extract_interp_tree(
            self.server_config.server_plugin_control.composite_forest,
            self.plugin_instance_id,
        )

    def load_config(
        self,
        plugin_config_dict,
    ) -> dict:
        return tree_path_interp_factory_config_desc.dict_from_input_dict(plugin_config_dict)

    def load_interp_tree_abs_paths(
        self,
        this_plugin_instance_interp_tree_abs_paths: list[tuple[str, ...]],
    ):
        # TODO: TODO_18_51_46_14: refactor FS_42_76_93_51 zero_arg_interp into FS_15_79_76_85 line processor:
        #       At the moment, `InterpTreeInterpFactory` is not plugged into main interp tree -
        #       instead, it is plugged by `FirstArgInterpFactory`.

        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.interp_tree,
            self.interp_selector_tree,
        )
        # Walk configured interp tree and call `load_interp_tree_abs_paths` with `interp_tree_abs_path` for each interp.
        all_interp_tree_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()
        for interp_factory_instance_id in all_interp_tree_abs_paths:
            assert_plugin_instance_id(
                self.server_config,
                interp_factory_instance_id,
                PluginType.InterpFactoryPlugin,
            )
            interp_factory: AbstractInterpFactory = self.server_config.interp_factories[interp_factory_instance_id]
            interp_factory.load_interp_tree_abs_paths([
                tuple(list_abs_path)
                for list_abs_path in all_interp_tree_abs_paths[interp_factory_instance_id]
            ])

    def load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> tuple[
        # mapped_func_ids:
        list[str],
        # index_props:
        list[str],
        # func_data_envelopes:
        list[dict],
    ]:
        with self._is_recursive_load() as is_recursive_load:
            # TODO: FS_33_76_82_84 composite forest: add validation that same interp tree interp is not loaded twice - is it required (given that plugin instances can be reused)?
            if is_recursive_load:
                return []
            return self._load_func_envelopes(
                interp_tree_abs_path,
                func_ids_to_func_envelopes,
            )

    @contextlib.contextmanager
    def _is_recursive_load(self):
        """
        Auto-reset flag `self.is_recursive_func_load` on exit from `with` section.
        """
        try:
            if not self.is_recursive_func_load:
                self.is_recursive_func_load = True
                is_recursive_load = False
            else:
                is_recursive_load = True
            yield is_recursive_load
        finally:
            self.is_recursive_func_load = False

    def _load_func_envelopes(
        self,
        interp_tree_abs_path: tuple[str, ...],
        func_ids_to_func_envelopes: dict[str, dict],
    ) -> tuple[
        # mapped_func_ids:
        list[str],
        # index_props:
        list[str],
        # func_data_envelopes:
        list[dict],
    ]:
        (
            mapped_func_ids,
            index_props,
            func_data_envelopes,
        ) = super().load_func_envelopes(
            interp_tree_abs_path,
            func_ids_to_func_envelopes,
        )
        dict_tree_walker: DictTreeWalker = DictTreeWalker(
            CompositeInfoType.interp_tree,
            self.interp_selector_tree,
        )
        # Walk configured interp tree and call `load_func_envelopes` with `interp_tree_abs_path` for each interp.
        all_interp_tree_abs_paths: dict[str, list[list[str]]] = dict_tree_walker.build_str_leaves_paths()
        for interp_factory_instance_id in all_interp_tree_abs_paths:
            for sub_interp_tree_abs_path in all_interp_tree_abs_paths[interp_factory_instance_id]:
                if not sequence_starts_with(sub_interp_tree_abs_path, interp_tree_abs_path):
                    # skip: other `interp_tree_abs_path`-s are going to be separate calls to this func:
                    continue
                assert_plugin_instance_id(
                    self.server_config,
                    interp_factory_instance_id,
                    PluginType.InterpFactoryPlugin,
                )
                interp_factory: AbstractInterpFactory = self.server_config.interp_factories[interp_factory_instance_id]

                (
                    extra_mapped_func_ids,
                    extra_index_props,
                    extra_func_data_envelopes,
                ) = interp_factory.load_func_envelopes(
                    tuple(sub_interp_tree_abs_path),
                    func_ids_to_func_envelopes,
                )
                mapped_func_ids.extend(extra_mapped_func_ids)
                index_props.extend(extra_index_props)
                func_data_envelopes.extend(extra_func_data_envelopes)
        return (
            mapped_func_ids,
            index_props,
            func_data_envelopes,
        )

    def create_interp(
        self,
        interp_ctx: InterpContext,
    ) -> InterpTreeInterp:
        """
        Each `InterpTreeInterpFactory` serves a subtree,
        but it only stores config of per its plug points
        (interp tree abs paths where it was invoked on `load_func_envelopes`).
        If next interp with its plugin id is created, but the specified path is within its subtree,
        chop off the last step in the path until it matches path to its subtree to select one.
        For example:
        ```
        # Current path within interp tree:
        interp_ctx.interp_tree_abs_path = ('l1', 'l2', 'l3', 'l4', )
        # Plug point of this factory within the interp tree:
        self.interp_tree_abs_paths_to_node_configs = {
            ('l1', 'l2', ): {},
            ('x1', 'x2', 'x3', ): {},
        }
        # Keep trying to cut of the last step in the path until it matches one of the key:
        ('l1', 'l2', 'l3', 'l4', )
        ('l1', 'l2', 'l3', )
        ('l1', 'l2', )
        # Done.
        ```
        """
        orig_subtree_abs_path = interp_ctx.interp_tree_abs_path
        curr_subtree_abs_path = orig_subtree_abs_path
        while curr_subtree_abs_path not in self.interp_tree_abs_paths_to_node_configs:
            if len(curr_subtree_abs_path) == 0:
                raise ValueError(
                    f"path `{orig_subtree_abs_path}` is not subpath of `{self.interp_tree_abs_paths_to_node_configs.keys()}` paths"
                )
            curr_subtree_abs_path = curr_subtree_abs_path[:-1]
        return InterpTreeInterp(
            self.plugin_instance_id,
            self.interp_tree_abs_paths_to_node_configs[curr_subtree_abs_path],
            interp_ctx,
            self.interp_selector_tree,
        )


class InterpTreeInterp(AbstractInterp):
    """
    Implements FS_01_89_09_24 interp tree.
    """

    def __init__(
        self,
        interp_factory_id: str,
        interp_tree_node_config_dict: dict,
        interp_ctx: InterpContext,
        interp_selector_tree: dict,
    ):
        super().__init__(
            interp_factory_id,
            interp_tree_node_config_dict,
            interp_ctx,
        )
        self.interp_selector_tree: dict = interp_selector_tree
        self.next_interp_factory_id: Union[str, None] = None

        # TODO: Why hard-coded? Isn't it possible for this plugin to be plugged into any depth of the tree?
        # Token with ipos = 0 is the command name eaten by `FirstArgInterp` (FS_42_76_93_51 first interp):
        self.base_token_ipos: int = 1

    def consumes_args_at_once(self) -> bool:
        return True

    def consume_dictated_args(self) -> bool:
        """
        As of now, `InterpTreeInterp` ignores `dictated_arg`-s.
        """
        return False

    def consume_offered_args(
        self,
    ) -> bool:
        """
        Consumes heading `offered_arg`-s according to the possible paths within `interp_selector_tree`.

        Unlike normal guideline to consume one `offered_arg`-s at time before re-query (FS_44_36_84_88),
        this func consumes all possible `offered_arg`-s at once
        because `arg_value`-s are selected by walking the `interp_selector_tree` (not via data query)
        and do not become incompatible by consuming all
        (do not violate FS_51_67_38_37 avoid impossible arg combinations).

        Also, FS_01_89_09_24 interp tree interp consumes only from
        the first (ipos = 0) `remaining_offered_args_per_bucket`.
        It still respects FS_97_64_39_94 `token_bucket` boundaries but
        does not have the same flexibility as consumption by FS_26_43_73_72 func tree interp.
        """

        curr_sub_tree = fetch_subtree_node(
            self.interp_selector_tree,
            self.interp_tree_abs_path,
        )
        any_consumed = False
        while True:
            if isinstance(curr_sub_tree, str):
                # Tree leaf is reached - use it:
                self.next_interp_factory_id = curr_sub_tree
                return any_consumed

            # Always consume next remaining `offered_arg`:
            offered_arg: Union[ArgCommandValueOffered, None] = self._next_remaining_offered_arg_from_first_bucket()

            if offered_arg is None:
                self.set_default_factory_id(curr_sub_tree)
                return any_consumed

            if isinstance(curr_sub_tree, dict):
                if offered_arg.get_arg_value() in curr_sub_tree:
                    # Consume one more arg into path:
                    self.interp_tree_abs_path.append(offered_arg.get_arg_value())

                    bucket_index = self.interp_ctx.token_ipos_to_token_bucket_map[offered_arg.get_arg_token()]
                    # Always consume from the first `token_bucket` (ipos = 0):
                    assert bucket_index == 0
                    # TODO: TODO_66_09_41_16: clarify command line processing
                    #       Use helper functions which ensure consistency
                    #       (like in this case, if one is removed, another is removed)
                    self.interp_ctx.consumed_token_buckets[bucket_index].append(offered_arg.get_arg_token())
                    self.interp_ctx.remaining_token_buckets[bucket_index].remove(offered_arg.get_arg_token())
                    self.interp_ctx.remaining_offered_args_per_bucket[bucket_index].remove(offered_arg)

                    curr_sub_tree = curr_sub_tree[offered_arg.get_arg_value()]
                    any_consumed = True
                    continue
                else:
                    self.set_default_factory_id(curr_sub_tree)
                    return any_consumed
            else:
                raise LookupError()

    def _next_remaining_offered_arg_from_first_bucket(
        self,
    ) -> Union[ArgCommandValueOffered, None]:
        for offered_arg in self.interp_ctx.remaining_offered_args_per_bucket[0]:
            return offered_arg
        return None

    def set_default_factory_id(
        self,
        curr_sub_tree,
    ):
        # Impossible to consume more arg - use default of the current sub-tree:
        if surrogate_node_id_ in curr_sub_tree:
            self.next_interp_factory_id = curr_sub_tree[surrogate_node_id_]
        else:
            # TODO: TODO_62_75_33_41: Do not hardcode plugin instance id (instance of `NoopInterpFactory`):
            self.next_interp_factory_id = f"{NoopInterpFactory.__name__}.default"

    def try_iterate(self) -> InterpStep:
        return InterpStep.NextInterp

    def next_interp(self) -> "AbstractInterp":
        # Selected `next_interp_factory_id` need to be given one of the paths in the FS_01_89_09_24 interp tree
        # (because each `next_interp_factory_id` can be plugged into multiple interp tree leaves).
        self.interp_ctx.interp_tree_abs_path = tuple(self.interp_tree_abs_path)
        return self.interp_ctx.create_next_interp(self.next_interp_factory_id)

    def propose_arg_completion(self) -> None:
        """
        Suggest based on current sub-tree from config.
        """

        if not self.is_eligible_for_suggestion():
            return

        curr_sub_tree = fetch_subtree_node(
            self.interp_selector_tree,
            self.interp_tree_abs_path,
        )
        if isinstance(curr_sub_tree, dict):
            for proposed_value in [
                x for x in curr_sub_tree
                if (
                    isinstance(x, str)
                    and
                    x != surrogate_node_id_
                    and
                    # FS_32_05_46_00: using `startswith`:
                    x.startswith(self.interp_ctx.parsed_ctx.tan_token_l_part)
                )
            ]:
                insert_unique_to_sorted_list(self.interp_ctx.comp_suggestions, proposed_value)

    def is_eligible_for_suggestion(self):
        """
        Suggesting anything is possible only if there are NO other tokens
        (after those recorded in `interp_tree_abs_path`)
        available for consumption by subsequent interpreters.
        """

        if self._next_remaining_offered_arg_from_first_bucket() is None:
            return True
        else:
            return False

        # TODO: TODO_66_09_41_16: clarify command line processing
        #       This comment and leftover code makes little sense now.
        #       Was it an attempt to implement suggestion when `tangent_token` is excluded and not eaten?
        #       So that we can suggest possible completions by interp tree interp?
        # TODO: TODO_51_14_50_19: ensure `tangent_token` always exists
        #       Clean up code below or take into account `tangent_token` (ipos cursor position).
        #       Note that there is no ipos cursor position at the moment.
        #       It has to be computed with help of FS_23_62_89_43 `tangent_token`
        #       (which might be a surrogate one if cursor does not touch non-whitespace chars).
        remaining_consumed_tokens = deepcopy(self.interp_ctx.consumed_token_ipos_list())

        # Remove everything until `base_token_ipos`:
        for token_ipos in range(0, self.base_token_ipos):
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)

        # Remove everything eaten into `interp_tree_abs_path`:
        token_ipos = self.base_token_ipos
        for node_path_token in self.interp_tree_abs_path:
            assert token_ipos in remaining_consumed_tokens
            remaining_consumed_tokens.remove(token_ipos)
            token_ipos += 1

        # Is there anything else?
        return len(remaining_consumed_tokens) == 0



from enum import (
    auto,
    Enum,
)


class ServicePropName(Enum):
    """
    Custom `prop_name`-s used by :class:`ServiceLoader`.

    See also:
    *   `TD_63_37_05_36` (demo) for `test_data`.
    """

    # ---

    code_maturity = auto()
    geo_region = auto()
    flow_stage = auto()

    # ---

    cluster_name = auto()

    # ---

    data_center = auto()
    host_name = auto()
    ip_address = auto()

    # FS_06_99_43_60: example of using non-scalar `prop_value` (array):
    group_label = auto()

    service_name = auto()

    run_mode = auto()

    # ---

    access_type = auto()

    # ---

    user_name = auto()

    # ---

    dir_path = auto()

    # ---

    live_status = auto()
    """
    A value manually or dynamically assigned to the resource indicating its status.
    """

    def __str__(self):
        return self.name



from marshmallow import (
    post_load,
    RAISE,
    Schema,
)

from argrelay_lib_root.misc_helper_common.TypeDesc import TypeDesc


class ObjectSchema(Schema):
    """
    Base schema for many other schemas which create object via @post_load.
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = dict

    @post_load
    def make_object(
        self,
        input_dict,
        **kwargs,
    ):
        """
        Implements inheritance as described here:
        https://stackoverflow.com/a/65668854/441652
        """
        return type(self).model_class(**input_dict)


object_desc = TypeDesc(
    dict_schema = ObjectSchema(),
    ref_name = ObjectSchema.__name__,
    dict_example = {},
    default_file_path = "",
)
